/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/callback/CallbackPollThread.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc. 
 * All rights reserved.
 */
package com.netapp.common.callback;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.apache.cxf.frontend.ClientProxyFactoryBean;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.api.OperationResult;
import com.netapp.common.flow.api.OperationService;
import com.netapp.common.flow.api.OperationState;
import com.netapp.common.flow.api.OperationUpdateTag;
import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.LogMessage;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ConfigBundle;

/**
 * A thread that will attempt to poll the remote server for information about a given set of ids.
 * 
 * @author joshb
 *
 */
public class CallbackPollThread extends Thread {
    private static final NALogger LOGGER = NALogger.getLogger(CallbackPollThread.class);

    protected boolean stopped;
    protected boolean autoRelease;
    protected long sleepTime;
    protected ServerEndPoint endpoint;
    protected List<IdCallbackPair> callbackIds;
    protected OperationUpdateTag lastUpdate;
    protected int maxSleepRetries;
    protected ConfigBundle configuration;
    protected OperationService operationService;

    public CallbackPollThread() {
        setCallbackIds(new Vector<IdCallbackPair>());
        setStopped(false);
    }

    public void run() {
        init();
        sleepUntilReady();

        int counter = 0;
        int maxRetry = CallbackConstants.MAX_POLL_RETRY.lookup(getConfiguration());

        while (!isStopped()) {
            try {
                // Sleep for up to N seconds
                sleep(getSleepTime() * 1000);
            }
            catch (InterruptedException e) {
            }

            // If there are no more ids to poll for, we are done.
            if (getCallbackIds().isEmpty()) {
                LOGGER.internal("Exiting due to no ids left");
                break;
            }

            String[] ids = new String[getCallbackIds().size()];

            for (int i = 0; i < getCallbackIds().size(); i++) {
                ids[i] = getCallbackIds().get(i).getId();
            }

            // make a request to the server for all of the IDs at once from the last update tag
            LOGGER.debug(MsgKey.CALLBACK_POLL_REQUEST, ids.length, getEndpoint().getHostname());

            try {
                OperationResult[] results = retrieveResults(ids);
                processResults(results);
            }
            catch (CallbackException e) {
                counter++;
                LOGGER.warn(MsgKey.CALLBACK_POLL_FAILED_TO_RETRIEVE_RESULTS, counter, maxRetry);

                if (counter >= maxRetry) {
                    break;
                }
            }
        }
        
        /*
         * If the thread was not told to stop and there are still Ids to process, then
         * the thread is terminating early. Tell the callback methods about the ids that
         * are remaining so the application can clean up properly
         */
        if (!isStopped() && !getCallbackIds().isEmpty()) {
            for (IdCallbackPair pair : getCallbackIds()) {
                pair.getCallback().unprocessedId(pair.getId());
            }
        }
    }

    /**
     * Basic thread initialization. This is the first thing that the thread will do.
     */
    protected void init() {
        setSleepTime(CallbackConstants.POLL_INTERVAL.lookup(getConfiguration()));
        setMaxSleepRetries(CallbackConstants.MAX_SETUP_POLL_RETRIES.lookup(getConfiguration()));
    }

    /**
     * Retrieve the operation results from the remote server through the OperationService SOAP object.
     * 
     * @param ids
     * @return Operation results for the set of passed in ids.
     * @throws CallbackException 
     */
    protected OperationResult[] retrieveResults(String[] ids) throws CallbackException {
        try {
            return getService().operationResults(ids, getLastUpdate());
        }
        // CHECKSTYLE:OFF
        catch (Throwable e) {
            // CHECKSTYLE:ON
            throw new CallbackException(e);
        }
    }

    /**
     * Process the OperationResult entries. Invoke the proper callback methods for each operation a result
     * was retrieved for.
     * 
     * If an operation has completed, remove it from the internal list.
     * 
     * @param results
     */
    protected void processResults(OperationResult[] results) {
        for (OperationResult result : results) {
            List<LogMessage> messages = result.getMessages();
            List<LogMessage> progressMessages = new ArrayList<LogMessage>();

            if (null != result.getMessages()) {
                // sort the messages into progress events and other log messages
                for (Iterator<LogMessage> it = messages.iterator(); it.hasNext();) {
                    LogMessage current = it.next();

                    if (current.getLevel().equals(LogLevel.PROGRESS)) {
                        progressMessages.add(current);
                        it.remove();
                    }
                }
            }

            IdCallbackPair pair = getCallbackForOperation(result.getId(), result.getClientRequestId());

            LOGGER.internal("Invoking callback for " + pair.getId() + " on callback " + pair.getCallback());

            if (null != pair) {
                if (null != messages && !messages.isEmpty()) {
                    pair.getCallback().message(pair.getId(), messages);
                }

                if (null != progressMessages && !progressMessages.isEmpty()) {
                    pair.getCallback().progress(pair.getId(), progressMessages);
                }

                pair.getCallback().status(pair.getId(), result.getState());
            }

            // set the lastUpdate entry to be the lowest, non-null value we have
            if (null != result.getLastUpdate()) {
                if (null == getLastUpdate()) {
                    setLastUpdate(result.getLastUpdate());
                }
                else if (result.getLastUpdate().getTag() < getLastUpdate().getTag()) {
                    setLastUpdate(result.getLastUpdate());
                }
            }

            /*
             * If the operation has completed (success or failure), then release it from the server
             * and remove it from our list of operations to retrieve data for provided that the client
             * wants us to release it automatically for them.
             */
            if (result.getState().equals(OperationState.COMPLETE) || result.getState().equals(OperationState.FAILED)
                    || result.getState().equals(OperationState.INCOMPLETE)) {
                if (isAutoRelease()) {
                    try {
                        getService().operationRelease(pair.getId());
                    }
                    catch (CallbackException e) {
                    }
                }

                removeCallbackEntry(pair.getId());
            }
        }
    }

    /**
     * Loop through all of the operation id/callback pairs to find the id/callback for the given operation.
     * 
     * @param id
     * @param clientRequestId
     * @return IdCallbackPair for the requested id or client request id.
     */
    protected IdCallbackPair getCallbackForOperation(String id, String clientRequestId) {
        IdCallbackPair callbackPair = null;

        for (IdCallbackPair pair : getCallbackIds()) {
            if (pair.getId().equals(id) || pair.getId().equals(clientRequestId)) {
                callbackPair = pair;
                break;
            }
        }

        if (null == callbackPair) {
            LOGGER.warn(MsgKey.CALLBACK_POLL_NO_CALLBACK_FOUND_FOR_OPERATION, id, clientRequestId);
        }

        return callbackPair;
    }

    /**
     * Keep going back to sleep until at least one operation is ready to process.
     * Do not sleep for more than maxSleepRetries times before just giving up.
     */
    protected void sleepUntilReady() {
        // Don't actually start up if we do not have any work
        int counter = 1;

        while (getCallbackIds().isEmpty()) {
            try {
                LOGGER.internal("Sleeping due to no callbacks to process yet");
                counter++;
                sleep(getSleepTime() * 1000);
            }
            catch (InterruptedException e) {
            }

            if (getMaxSleepRetries() == counter) {
                setStopped(true);
                break;
            }
        }
    }

    /**
     * Remove a given ID so that it is no longer polled.
     * 
     * @param id
     */
    public void removeCallbackEntry(String id) {
        for (Iterator<IdCallbackPair> it = getCallbackIds().iterator(); it.hasNext(); ) {
            IdCallbackPair pair = it.next();
            
            if (pair.getId().equals(id)) {
                it.remove();

                LOGGER.debug(MsgKey.CALLBACK_POLL_REMOVE_OPERATION, id);
            }
        }
    }

    /**
     * Retrieve the location of the OperationService SOAP interface. Uses a proxy object to handle the
     * SOAP calls.
     * 
     * @return OperationService interface for the current endpoint.
     * @throws CallbackException 
     */
    protected OperationService getService() throws CallbackException {
        if (null != getOperationService()) {
            return getOperationService();
        }

        // How can I update this for IPv6?
        String url = getEndpoint().getProtocol() + "://" + getEndpoint().getHostname() + ":" + getEndpoint().getPort()
            + "/" + getEndpoint().getPath();

        LOGGER.debug(MsgKey.CALLBACK_POLL_CONNECTING_TO_OPERATION_SERVICE, url);

        try {
            ClientProxyFactoryBean factory = new ClientProxyFactoryBean();

            factory.setServiceClass(OperationService.class);
            factory.setAddress(url);

            setOperationService((OperationService) factory.create());
        }
        // CHECKSTYLE:OFF
        catch (Throwable t) {
            // CHECKSTYLE:ON
            LOGGER.debug(MsgKey.CALLBACK_POLL_CONNECT_TO_OPERATION_SERVICE_FAILED, t, url);
            throw new CallbackException(t);
        }

        return getOperationService();
    }

    public boolean isStopped() {
        return stopped;
    }

    public void setStopped(boolean stopped) {
        this.stopped = stopped;
    }

    protected long getSleepTime() {
        return sleepTime;
    }

    protected void setSleepTime(long sleepTime) {
        this.sleepTime = sleepTime;
    }

    public ServerEndPoint getEndpoint() {
        return endpoint;
    }

    public void setEndpoint(ServerEndPoint endpoint) {
        this.endpoint = endpoint;
    }

    protected List<IdCallbackPair> getCallbackIds() {
        return callbackIds;
    }

    protected void setCallbackIds(List<IdCallbackPair> callbackIds) {
        this.callbackIds = callbackIds;
    }

    public void addCallbackId(String id, OperationCallback callback) {
        getCallbackIds().add(new IdCallbackPair(id, callback));
    }

    public OperationUpdateTag getLastUpdate() {
        if (null == lastUpdate) {
            lastUpdate = new OperationUpdateTag();
        }

        return lastUpdate;
    }

    protected void setLastUpdate(OperationUpdateTag lastUpdate) {
        LOGGER.internal("Updating lastUpdate from %s to %s", this.lastUpdate, lastUpdate);
        this.lastUpdate = lastUpdate;
    }

    protected int getMaxSleepRetries() {
        return maxSleepRetries;
    }

    protected void setMaxSleepRetries(int maxSleepRetries) {
        this.maxSleepRetries = maxSleepRetries;
    }

    protected ConfigBundle getConfiguration() {
        return configuration;
    }

    protected void setConfiguration(ConfigBundle configuration) {
        this.configuration = configuration;
    }

    protected OperationService getOperationService() {
        return operationService;
    }

    protected void setOperationService(OperationService operationService) {
        this.operationService = operationService;
    }

    protected boolean isAutoRelease() {
        return autoRelease;
    }

    public void setAutoRelease(boolean autoRelease) {
        this.autoRelease = autoRelease;
    }

}
