/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/cli/AbstractCliGenerator.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.cli;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;

import com.netapp.common.MsgKey;
import com.netapp.common.cli.CliParameter.Type;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;
import com.netapp.common.util.StringUtils;

public abstract class AbstractCliGenerator {
    private static final NALogger LOGGER = NALogger.getLogger(AbstractCliGenerator.class);
    
    private static final ConfigProperty<String> ARGUMENT_SEPARATOR = ConfigProperty.stringProperty("cli.argument.separator", " ");
    private static final ConfigProperty<String> PARAMETER_PREFEX = ConfigProperty.stringProperty("cli.parameter.prefix", "-");
    private static final ConfigProperty<String> QUOTE_CHARACTER = ConfigProperty.stringProperty("cli.character.quote", "\"");
    private static final ConfigProperty<String> ESCAPE_CHARACTER = ConfigProperty.stringProperty("cli.character.escape");
    
    /*
     * The list of quotable and escapables must be separated by a comma.
     * 
     * Notes:
     *  * If you are escaping your escape character, put it first
     *  * We are performing a regular expression replacement, so escape characters as needed with a regex escape (\).
     */
    private static final ConfigProperty<String> QUOTABLES = ConfigProperty.stringProperty("cli.characters.quotables");
    private static final ConfigProperty<String> ESCAPABLES = ConfigProperty.stringProperty("cli.characters.escapables");
    

    protected String cliCommand;
    protected ConfigBundle context;
    protected String separator;
    protected String prefix;
    protected String escapeCharacter;
    protected String quoteCharacter;
    protected String [] escapables;
    protected String [] quotables;

    /**
     * Given an annotated method and a list of the arguments for that method, recreate what the
     * CLI would have been. 
     * 
     * Ensure that the list of arguments are the same number and type (in order) as those expected
     * by the method.
     * 
     * @param method
     * @param args
     * @throws CliGeneratorException 
     */
    protected void generateCliForMethod(Method method, Object... args) throws CliGeneratorException {
        Class<?>[] methodParameters = method.getParameterTypes();
        Annotation[][] parameterAnnotations = method.getParameterAnnotations();

        if (null == method) {
            throw new CliGeneratorException(LOGGER.error(MsgKey.CLI_GENERATOR_ERROR_NULL_METHOD));
        }

        // must have same number of arguments
        if (methodParameters.length != args.length) {
            throw new CliGeneratorException(LOGGER.error(MsgKey.CLI_GENERATOR_ERROR_PARAMETER_LENGTH,
                methodParameters.length, args.length));
        }

        // every argument must have an annotation
        if (methodParameters.length != parameterAnnotations.length) {
            throw new CliGeneratorException(LOGGER.error(MsgKey.CLI_GENERATOR_ERROR_PARAMETER_NOT_ANNOTATED, 
                method.getName(), methodParameters.length, parameterAnnotations.length));
        }

        // must have class types in the exact same order
        for (int i = 0; i < methodParameters.length; i++) {

            if (!ClassUtils.typeAssignableFrom(methodParameters[i], args[i].getClass())) {
                throw new CliGeneratorException(LOGGER.error(MsgKey.CLI_GENERATOR_ERROR_INVALID_CLASS_FOR_PARAMETER,
                    i, methodParameters[i].getName(), args[i].getClass().getName()));
            }

            if (null == getCliParameter(parameterAnnotations[i])) {
                throw new CliGeneratorException(LOGGER.error(MsgKey.CLI_GENERATOR_ERROR_PARAMETER_MISSING_ANNOTATION, i));
            }
        }

        Annotation[] methodAnnotations = method.getAnnotations();
        CliCommand methodCliCommandAnnotation = getCliCommand(methodAnnotations);

        if (null == methodCliCommandAnnotation) {
            throw new CliGeneratorException(LOGGER.error(MsgKey.CLI_GENERATOR_ERROR_METHOD_MISSING_ANNOTATION, method
                .getName()));
        }

        // Start building up the CLI
        LOGGER.internal("Validation complete. Time to build the CLI.");
        StringBuffer output = new StringBuffer(64);
        
        setSeparator(ARGUMENT_SEPARATOR.lookup(getContext()));
        setPrefix(PARAMETER_PREFEX.lookup(getContext()));
        setEscapeCharacter(ESCAPE_CHARACTER.lookup(getContext()));
        setQuoteCharacter(QUOTE_CHARACTER.lookup(getContext()));
        
        String temp = ESCAPABLES.lookup(getContext());
        
        if (!StringUtils.isEmpty(temp)) {
            setEscapables(temp.split(","));
        }
        else {
            setEscapables(new String[] {});
        }
        
        temp = QUOTABLES.lookup(getContext());
        
        if (!StringUtils.isEmpty(temp)) {
            setQuotables(temp.split(","));
        }
        else {
            setQuotables(new String[] {});
        }
        
        LOGGER.internal("Escape character is '%s'", getEscapeCharacter());
        LOGGER.internal("Escapable characters are '%s'", getEscapables().toString());

        /*
         * Add the command and potentially subcommand first
         */
        output.append(methodCliCommandAnnotation.command());
        output.append(getSeparator());

        if (!StringUtils.isEmpty(methodCliCommandAnnotation.subCommand())) {
            output.append(methodCliCommandAnnotation.subCommand());
            output.append(getSeparator());
        }

        /*
         * For each parameter to the method, add it
         */
        for (int i = 0; i < parameterAnnotations.length; i++) {
            output.append(processMethodArgument(parameterAnnotations[i], args[i]));
        }

        setCliCommand(output.toString());
    }

    /**
     * Look for the CliParameter annotations for the method. Using those annotations,
     * build the CLI for Bean, Array and primitive data types.
     * 
     * @param annotations
     * @param argument
     * @return cli for the supplied object and arguments
     * @throws CliGeneratorException 
     */
    protected String processMethodArgument(Annotation[] annotations, Object argument) throws CliGeneratorException {
        StringBuffer buffer = new StringBuffer(64);

        CliParameter parameter = getCliParameter(annotations);

        /*
         * Beans (Objects) require that we process all internal methods that contain a
         * CliParameter. Each of these methods can only take a single parameter and must
         * be a setter method. 
         */
        if (parameter.type().equals(Type.BEAN)) {
            buffer.append(getPrefix());
            buffer.append(parameter.option());
            buffer.append(getSeparator());

            buffer.append(processBean(argument));
        }
        /*
         * Expand an array to all of its values. Use the long version of the CLI
         * where every array element has the parameter switch in front of it.
         * If an array is a bean, expand the bean definition for each array element.
         */
        else if (argument.getClass().isArray()) {
            int length = Array.getLength(argument);

            for (int index = 0; index < length; index++) {
                buffer.append(getPrefix());
                buffer.append(parameter.option());
                buffer.append(getSeparator());

                Object value = Array.get(argument, index);

                if (isSimpleBean(value.getClass())) {
                    buffer.append(escape(value.toString()));
                    buffer.append(getSeparator());
                }
                else {
                    // recurse on a beans methods
                    buffer.append(processBean(value));
                }
            }
        }
        /*
         * Basic, primitive types require only a single parameter switch and value
         */
        else {
            /*
             * for boolean arguments, if the boolean value is false, we do not
             * display the option. If the boolean value is true, we do not
             * display the argument.
             */
            if ((!isBoolean(argument) && isDisplayable(argument)) || getBooleanValue(argument)) {
                buffer.append(getPrefix());
                buffer.append(parameter.option());
                buffer.append(getSeparator());
            
                if (!isBoolean(argument)) {
                    buffer.append(escape(argument.toString()));
                    buffer.append(getSeparator());
                }
            }
        }

        return buffer.toString();
    }

    /**
     * Build the CLI definition of a bean. If a bean method/value is not
     * a simple data type, then expand the bean definition. 
     * 
     * @param argument
     * @return cli definition for argument
     * @throws CliGeneratorException 
     */
    protected String processBean(Object argument) throws CliGeneratorException {
        StringBuffer buffer = new StringBuffer();

        Method[] beanMethods = argument.getClass().getMethods();

        for (Method method : beanMethods) {
            if (!ClassUtils.isBeanMethod(method)) {
                LOGGER.internal("Ignoring non-bean method");
                continue;
            }

            CliParameter beanParameter = getCliParameter(method.getAnnotations());

            if (null != beanParameter) {

                // retrieve the corresponding getter method of the object 'argument'
                String fieldName = ClassUtils.extractFieldNameFromBeanMethod(method);
                Method getterMethod = ClassUtils.findGetterMethod(argument.getClass(), fieldName);

                try {
                    Object value = getterMethod.invoke(argument);

                    if (isSimpleBean(value.getClass())) {
                        if ((!isBoolean(argument) && isDisplayable(argument)) || getBooleanValue(argument)) {
                            buffer.append(getPrefix());
                            buffer.append(beanParameter.option());
                            buffer.append(getSeparator());
                        }
                        
                        if (!isBoolean(argument)) {
                            buffer.append(escape(value.toString()));
                            buffer.append(getSeparator());
                        }
                    }
                    else {
                        List<Method> setterMethods = ClassUtils.findSetterMethod(argument.getClass(), fieldName);

                        if (0 == setterMethods.size()) {
                            throw new CliGeneratorException(LOGGER.error(MsgKey.CLI_GENERATOR_ERROR_MISSING_SETTER_METHOD, fieldName));
                        }
                        else if (1 < setterMethods.size()) {
                            LOGGER.warn(MsgKey.CLI_GENERATOR_ERROR_TOO_MANY_SETTER_METHODS, setterMethods.size(), fieldName);
                        }
                        
                        // recurse with the annotations and the object
                        buffer.append(processMethodArgument(setterMethods.get(0).getAnnotations(), value));
                    }
                }
                catch (IllegalArgumentException e) {
                    throw new CliGeneratorException(e);
                }
                catch (IllegalAccessException e) {
                    throw new CliGeneratorException(e);
                }
                catch (InvocationTargetException e) {
                    throw new CliGeneratorException(e);
                }
            }
        }

        return buffer.toString();
    }

    /**
     * A bean is a simple bean if none of its methods contain a CliParameter annotation.
     * Or, to rephrase it. A bean is not a simple bean if one method contains a CliParameter
     * annotation.
     * 
     * Additionally, a bean is NOT simple if it is an array.
     * 
     * @param clazz
     * @return true if clazz represents a simple bean
     */
    protected boolean isSimpleBean(Class<?> clazz) {
        /*
         * Special case. Arrays are NOT simple.
         */
        if (clazz.isArray()) {
            return false;
        }

        Method[] methods = clazz.getMethods();

        for (Method method : methods) {
            CliParameter cliParameter = getCliParameter(method.getAnnotations());

            /*
             * If we do find a single CliParameter, this bean is not simple
             */
            if (null != cliParameter) {
                return false;
            }
        }

        return true;
    }
    
    /**
     * An argument is displayable if it is not null or an empty value.
     * 
     * @param argument
     * @return true if argument is displayable
     */
    protected boolean isDisplayable(Object argument) {
        boolean displayable = false;
        
        if (null != argument) {
            if (argument instanceof String) {
                String temp = (String) argument;
                
                if (!StringUtils.isBlank(temp)) {
                    displayable = true;
                }
            }
            else {
                displayable = true;
            }
        }
    
        return displayable;
    }
    
    /**
     * Check if the supplied object is a boolean value.
     * 
     * @param argument
     * @return true if the supplied argument is a Boolean or boolean type.
     */
    protected boolean isBoolean(Object argument) {
        if (argument.getClass().equals(Boolean.class) || argument.getClass().equals(Boolean.TYPE)) {
            return true;
        }
        
        return false;
    }
    
    /**
     * Check if the supplied argument is a boolean and return the boolean value.
     * 
     * @param argument
     * @return value of argument if it can be interpreted as a boolean value, false otherwise.
     */
    protected boolean getBooleanValue(Object argument) {
        if (!isBoolean(argument)) {
            return false;
        }
        
        return (Boolean) argument;
    }
    
    /**
     * If the original string contains a character which requires the string to be quoted, quote
     * the string once and only once.
     * 
     * For every occurance of an escapable character in the original string, add the
     * escape character in from of it.
     * 
     * @param original
     * @return escaped and quoted version of original
     */
    protected String escape(String original) {
        for (String quotable : getQuotables()) {
            if (original.contains(quotable)) {
                original = getQuoteCharacter() + original + getQuoteCharacter();
                break; // quote once and only once
            }
        }
        
        for (String escapble : getEscapables()) {
            original = original.replaceAll(escapble, getEscapeCharacter() + escapble);
        }
        
        return original;
    }

    protected CliParameter getCliParameter(Annotation[] annotations) {
        for (Annotation annotation : annotations) {
            if (annotation.annotationType().equals(CliParameter.class)) {
                return (CliParameter) annotation;
            }
        }

        return null;
    }

    protected CliCommand getCliCommand(Annotation[] annotations) {
        for (Annotation annotation : annotations) {
            if (annotation.annotationType().equals(CliCommand.class)) {
                return (CliCommand) annotation;
            }
        }

        return null;
    }

    public String getCliCommand() {
        return cliCommand;
    }

    protected void setCliCommand(String cliCommand) {
        this.cliCommand = cliCommand;
    }

    protected ConfigBundle getContext() {
        if (null == context) {
            context = new ConfigBundle();
        }
        
        return context;
    }

    public void setContext(ConfigBundle context) {
        this.context = context;
    }

    protected String getSeparator() {
        return separator;
    }

    protected void setSeparator(String separator) {
        this.separator = separator;
    }

    protected String getPrefix() {
        return prefix;
    }

    protected void setPrefix(String prefix) {
        this.prefix = prefix;
    }

    protected String getEscapeCharacter() {
        return escapeCharacter;
    }

    protected void setEscapeCharacter(String escapeCharacter) {
        this.escapeCharacter = escapeCharacter;
    }

    protected String[] getEscapables() {
        return escapables;
    }

    protected void setEscapables(String[] escapables) {
        this.escapables = escapables;
    }

    protected String getQuoteCharacter() {
        return quoteCharacter;
    }

    protected void setQuoteCharacter(String quoteCharacter) {
        this.quoteCharacter = quoteCharacter;
    }

    protected String[] getQuotables() {
        return quotables;
    }

    protected void setQuotables(String[] quotables) {
        this.quotables = quotables;
    }
}
