/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/execution/ShellResult.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.execution;

import static com.netapp.common.MsgKey.*;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.lang.time.DurationFormatUtils;
import org.apache.commons.lang.time.StopWatch;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.PlatformConfigBundle;
import com.netapp.common.util.CollectionUtils;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;
import com.netapp.common.util.Utils;

/**
 * An object representing the result of an executed shell command.
 * Holds the STDOUT, STDERR and exit code of the process.
 * Provides methods to wait for the command to complete.
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class ShellResult implements Serializable {
    private static final long serialVersionUID = 1L;
    private static final NALogger logger = NALogger.getLogger(ShellResult.class);
    
    private ShellRunnable commandExecuted;
    
    private List<String> stdErr;
    private List<String> stdOut;
    private List<String> allOut;
    
    private transient ProcessStreamReadingThread stdErrThread;
    private transient ProcessStreamReadingThread stdOutThread;
    private transient Thread executionThread;

    private transient ProcessContext processContext;
    
    private static final ConfigProperty<Long> DEFAULT_TIMEOUT_PROPERTY = ConfigProperty.longProperty("shellResult.timeout", 365 * 24 * 60 * 60 * 1000L); // 1 year 

    private static final ConfigBundle CONFIG_BUNDLE = new PlatformConfigBundle();
    
    public static final ProcessListener NULL_LISTENER = new ProcessListener() {
    
        public void handleStdOut(ProcessContext processContext, String line) {
        }
    
        public void handleStdErr(ProcessContext processContext, String line) {
        }
    };
    
    private abstract class ProcessStreamReadingThread extends Thread {
        
        public ProcessStreamReadingThread(String name) {
            super(name);
        }
        
        public void run() {
            BufferedReader read = null;
            try {
                read = new BufferedReader(new InputStreamReader(getStreamToRead()));
                // Break out of the loop if this thread has been interrupted
                while (!Thread.interrupted()) {
                    /*
                     *  Still a small window here where we can call readLine on a closed stream
                     *  (if the thread gets interrupted after we've checked the flag)
                     *  Oh well.
                     */ 
                    String line = read.readLine();
                    if (line == null) {
                        break;
                    }
                    getLineCollection().add(line);
                    ShellResult.this.allOut.add(line);
                    invokeProcessListener(line);
                }
            }
            /*
             * CHECKSTYLE:OFF
             * Catch base Exception class, because this thread can throw an exception directly
             * to the user, which we want to prevent
             */
            catch (Throwable t) {
                // CHECKSTYLE:ON
                if (!Thread.interrupted()) {
                    logger.error(getErrorMessageKey(), t, 
                        ShellResult.this.commandExecuted.getUserFacingCommand(), 
                        Utils.prettyString(t));
                    
                    /*
                     * burt 318646
                     * 
                     * If an exception occurs while handling stdout or stderr,
                     * the ProcessListener should do its best job to shutdown the process gracefully.
                     * 
                     * After the ProcessListener attempts to shutdown the process gracefully,
                     * we kill the process here to keep from hanging forever if the process
                     * is waiting for a response.
                     * 
                     * Also, we swallow the exception because we expect the user of ShellResult to
                     * handle the output or exit code.
                     */
                    logger.error(EXECUTION_KILLING_PROCESS, ShellResult.this.commandExecuted.getUserFacingCommand());
                    ShellResult.this.processContext.getProcess().destroy();
                }
            }
            finally {
                if (read != null) {
                    try {
                        read.close();
                    }
                    catch (IOException e) {
                        // Ignore
                    }
                }
            }
        }

        protected abstract InputStream getStreamToRead();
        
        protected abstract List<String> getLineCollection();
        
        protected abstract void invokeProcessListener(String line);

        protected abstract MsgKey getErrorMessageKey();

    }
    
    /**
     * Construct from a live process
     */
    public ShellResult(ProcessContext processContext) {
        this(processContext, NULL_LISTENER);
    }

    /**
     * Testing / Override hook
     */
    protected ShellResult(List<String> stdOut, List<String> stdErr, List<String> allOut, ShellRunnable commandExecuted) {
        this.stdOut = stdOut;
        this.stdErr = stdErr;
        this.allOut = allOut;
        this.commandExecuted = commandExecuted;
    }

    public ShellResult(ProcessContext processContext, final ProcessListener processListener) {
        this.processContext = processContext;
        this.commandExecuted = this.processContext.getShellCommand();
        this.stdErr = Collections.synchronizedList(new LinkedList<String>());
        this.stdOut = Collections.synchronizedList(new LinkedList<String>());
        this.allOut = Collections.synchronizedList(new LinkedList<String>());

        this.stdErrThread = new ProcessStreamReadingThread(String.format("STDERR Capture Thread [%s]", this.commandExecuted)) {
            
            @Override
            protected InputStream getStreamToRead() {
                return ShellResult.this.processContext.getProcess().getErrorStream();
            }

            @Override
            protected List<String> getLineCollection() {
                return ShellResult.this.stdErr;
            }

            @Override
            protected void invokeProcessListener(String line) {
                processListener.handleStdErr(ShellResult.this.processContext, line);
            }
            
            @Override
            protected MsgKey getErrorMessageKey() {
                return EXECUTION_CANNOT_READ_ERROR_STREAM;
            }
            
        };

        this.stdOutThread = new ProcessStreamReadingThread(String.format("STDOUT Capture Thread [%s]", this.commandExecuted)) {
            
            @Override
            protected InputStream getStreamToRead() {
                return ShellResult.this.processContext.getProcess().getInputStream();
            }

            @Override
            protected List<String> getLineCollection() {
                return ShellResult.this.stdOut;
            }

            @Override
            protected void invokeProcessListener(String line) {
                processListener.handleStdOut(ShellResult.this.processContext, line);
            }
            
            @Override
            protected MsgKey getErrorMessageKey() {
                return EXECUTION_CANNOT_READ_INPUT_STREAM;
            }

        };

        this.executionThread = new Thread(String.format("Execution Monitor Thread [%s]", this.commandExecuted)) {
            @Override
            public void run() {
                try {
                    ShellResult.this.processContext.getProcess().waitFor();
                    ShellResult.this.stdOutThread.join();
                    ShellResult.this.stdErrThread.join();
                }
                catch (InterruptedException e) {
                    logger.error(EXECUTION_KILLING_PROCESS, ShellResult.this.commandExecuted.getUserFacingCommand());
                    ShellResult.this.stdErrThread.interrupt();
                    ShellResult.this.stdOutThread.interrupt();
                    ShellResult.this.processContext.getProcess().destroy();
                }
                finally {
                    StopWatch timer = ShellResult.this.processContext.getTimer();
                    timer.stop();
                    
                    logger.debug(EXECUTION_SHELL_RESULT, timer.toString(), ShellResult.this.getExitValue(), 
                        CollectionUtils.collectionToString(ShellResult.this.allOut));
                }   
            }
        };
        
        this.stdErrThread.start();
        this.stdOutThread.start();
        this.executionThread.start();
    }

    /**
     * Wait for the process to finish executing (uses a default timeout).
     */
    public ShellResult waitFor() {
        return this.waitFor(DEFAULT_TIMEOUT_PROPERTY.lookup(CONFIG_BUNDLE));
    }

    /**
     * Wait for the process to finish executing.
     * 
     * @param timeoutInMsec Duration to wait before killing the process and returning
     */
    public ShellResult waitFor(long timeoutInMsec) {
        try {
            this.executionThread.join(timeoutInMsec);
        }
        catch (InterruptedException e) {
            // No-op.  This is the parent thread.
        }
        
        if (this.executionThread.isAlive()) {
            /*
             * The executionThread.join() call has timed out before the executionThread has finished 
             * --> Interrupt the execution thread (terminate the process)
             */
            this.executionThread.interrupt();
            try {
                this.executionThread.interrupt();
                this.executionThread.join();
            }
            catch (InterruptedException e) {
                // No-op.  This is the parent thread.
            }

            throw new ProcessExecutionException(logger.error(EXECUTION_TIMEOUT_OCCURED, 
                DurationFormatUtils.formatDurationHMS(timeoutInMsec), 
                this.processContext.getShellCommand().getUserFacingCommand()));
        }
        
        return this;
    }
    
    public void destroy() {
        try {
            this.executionThread.interrupt();
            this.executionThread.join();
        }
        catch (InterruptedException e) {
            // No-op.  This is the parent thread.
        }
    }

    public List<String> getAllOut() {
        return this.allOut;
    }

    public String allOutAsString() {
        return CollectionUtils.collectionToString(getAllOut());
    }
    
    public Integer getExitValue() {
        try {
            return this.processContext.getProcess().exitValue();
        }
        catch (IllegalThreadStateException e) {
            return null;
        }
    }

    public List<String> getStdErr() {
        return this.stdErr;
    }
    
    public String stdErrAsString() {
        return CollectionUtils.collectionToString(getStdErr());
    }
    
    public List<String> getStdOut() {
        return this.stdOut;
    }

    public String stdOutAsString() {
        return CollectionUtils.collectionToString(getStdOut());
    }
    
    public ShellRunnable getCommandExecuted() {
        return this.commandExecuted;
    }
    
    public boolean isZeroExitCode() {
        return this.getExitValue() == 0;
    }
    
    public boolean isNonZeroExitCode() {
        return !this.isZeroExitCode();
    }
    
    public boolean isRunning() {
        /*
         * exitValue() will be null if the process is still running
         */
        return this.getExitValue() == null;
    }
    
    @Override
    public String toString() {
        return CollectionUtils.collectionToString(this.allOut);
    }

    public ProcessContext getProcessContext() {
        return this.processContext;
    }
}
