/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/api/OperationResult.java#1 $
 *
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow.api;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.LogMessage;
import com.netapp.common.logging.NALogger;

/**
 * Base OperationResult for SOAP interactions.
 * Operations with additonal outputs should subclass this and
 * add fields, getters, and setters as required to match the register
 * definitions.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
@XmlRootElement(name = "result")
@XmlType(name = "operationResultType", namespace = "http://www.netapp.com/smai/Operation")
public class OperationResult implements Serializable, Cloneable {
    @SuppressWarnings("unused")
    private static final NALogger logger = NALogger.getLogger(OperationResult.class);
    /**
     *
     */
    private static final long serialVersionUID = 1L;
    /**
     * Workflow engine id of this operation.
     */
    protected String id;

    /**
     * Client tracking identifier of this operation, or null if no id was supplied.
     */
    protected String clientRequestId;

    /**
     * Operation type of this operation.
     */
    protected String type;

    /**
     * Current state of this operation.
     */
    protected OperationState state;

    /**
     * Last checkpoint time of this operation.
     */
    protected Date lastActiveTime;

    /**
     * Last time a log message or the state changed.
     */
    protected OperationUpdateTag lastUpdate;

    /**
     * List of logged messages for this operation.
     * Map of output values for this operation. Only filled in
     * if {@link #state} is {@link OperationState#COMPLETE COMPLETE},
     * {@link OperationState#FAILED FAILED} or
     * {@link OperationState#INCOMPLETE INCOMPLETE}.
     */
    protected List<LogMessage> messages;

    public OperationResult() {

    }

    public void setId(String id) {
        this.id = id;
    }
    /**
     * Retrieve the id of this operation
     * @return operation id
     */
    public String getId() {
        return this.id;
    }

    public void setClientRequestId(String clientRequestId) {
        this.clientRequestId = clientRequestId;
    }

    public String getClientRequestId() {
        return this.clientRequestId;
    }

    public void setType(String type) {
        this.type = type;
    }

    /**
     * Get the operation type for this operation.
     * @return Operation type of the operation.
     */
    public String getType() {
        return this.type;
    }

    public void setLastActiveTime(Date lastActiveTime) {
        this.lastActiveTime = lastActiveTime;
    }

    /**
     * Retrieve the last time this operation checkpointed.
     * @return operation checkpoint tme
     */
    public Date getLastActiveTime() {
        return this.lastActiveTime;
    }

    public void setState(OperationState state) {
        this.state = state;
    }

    /**
     * Retrieve the state of this operation.
     * If the operation is not yet terminal, the state is always exposed
     * as RUNNING.
     * @return terminal operation state
     */
    public OperationState getState() {
        return this.state;
    }

    public void setMessages(List<LogMessage> messages) {
        this.messages = messages;
    }

    /**
     * Retrieve the messages logged for this operation.
     * @return List of logged messages associated to this operation
     */
    public List<LogMessage> getMessages() {
        return this.messages;
    }
    
    /**
     * @return The first message logged for this operation with level ERROR or FATAL or null if no such
     * message is found.
     */
    public LogMessage getFirstError() {
        for (LogMessage message : this.messages) {
            if (LogLevel.ERROR.equals(message.getLevel()) || LogLevel.FATAL.equals(message.getLevel())) {
                return message;
            }
        }
        
        return null;
    }

    /**
     * Retrieve the opaque object for the last update.
     * @return last update tag
     */
    public OperationUpdateTag getLastUpdate() {
        return lastUpdate;
    }

    public void setLastUpdate(OperationUpdateTag lastUpdate) {
        this.lastUpdate = lastUpdate;
    }

    @Override
    public int hashCode() {
        return this.id.hashCode();
    }

    @Override
    public boolean equals(Object o) {
        return ((o instanceof OperationResult) && ((OperationResult)o).id.equals(this.id));
    }

    @Override
    public Object clone() {
        try {
            OperationResult clone = (OperationResult) super.clone();

            // cloning these two items is not quite straight forward. java.util.Date might already
            // work correctly, but I'm just making sure.
            if (null != getLastUpdate()) {
                clone.setLastUpdate(new OperationUpdateTag(getLastUpdate().getTag()));
            }

            if (null != getLastActiveTime()) {
                clone.setLastActiveTime(new Date(getLastActiveTime().getTime()));
            }

            clone.setMessages(new ArrayList<LogMessage>());

            if (null != getMessages()) {
                for (LogMessage message : getMessages()) {
                    // not a deep clone of LogMessage
                    clone.getMessages().add(message);
                }
            }

            return clone;
        }
        catch (CloneNotSupportedException e) {
            throw new RuntimeException(e);
        }
    }
}