/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/soap/OperationServiceImpl.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow.soap;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Date;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.Engine;
import com.netapp.common.flow.Result;
import com.netapp.common.flow.api.OperationResult;
import com.netapp.common.flow.api.OperationService;
import com.netapp.common.flow.api.OperationState;
import com.netapp.common.flow.api.OperationUpdateTag;
import com.netapp.common.logging.LogMessage;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;

/**
 * Implementation of the {@link OperationService} web service.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class OperationServiceImpl implements OperationService {
    @SuppressWarnings("unused")
    private static final NALogger logger = NALogger.getLogger(OperationServiceImpl.class);
    
    private Engine engine;
    private ResultDelegate resultDelegate;
    
    public OperationServiceImpl() {
    }

    public OperationServiceImpl(Engine engine) {
        this.engine = engine;
    }

    public void setEngine(Engine engine) {
        assert this.engine == null;
        this.engine = engine;
    }

    public void setResultDelegate(ResultDelegate resultDelegate) {
        this.resultDelegate = resultDelegate;
    }
    /**
     * 
     * @see com.netapp.common.flow.api.OperationService#operationCancel(java.lang.String)
     */
    public void operationCancel(String id) {
        assert this.engine != null;
        engine.operationCancel(id);
    }

    public void operationRelease(String id) {
        assert this.engine != null;
        engine.operationRelease(id);
    }

    /**
     * 
     * @see com.netapp.common.flow.api.OperationService#operationResult(java.lang.String)
     */
    public OperationResult operationResult(String id) {
        assert this.engine != null;
        return operationResult(id, OperationResult.class);
    }

    /**
     * @see com.netapp.common.flow.api.OperationService#operationResults(java.lang.String[], com.netapp.common.flow.api.OperationUpdateTag)
     */
    public OperationResult[] operationResults(String[] ids, OperationUpdateTag lastUpdate) {
        assert this.engine != null;

        OperationResult[] results = new OperationResult[ids.length];

        for (int i = 0; i < ids.length; i++) {
            results[i] = trimOldMessages(operationResult(ids[i], OperationResult.class), lastUpdate);
        }

        return results;
    }

    /**
     * Construct an extended Operation Result. This is used by {@link WorkflowBindingFactory} to
     * implement result calls that return outputs back to the client.
     * <ul>
     *   <li>The workflow result for <samp>id</samp> is retrieved.
     *   <li>A class of <samp>T</samp> is constructed via the no-argument public constructor.
     *   <li>Any setFoo methods in <samp>T</samp> with a corresponding getFoo method in 
     *       {@link com.netapp.common.flow.Result} are populated from the base operation result.
     *   <li>If the operation is complete, any remaining setFoo methods in <samp>T</samp> are set from the
     *       value in the result output map.
     * </ul>
     * @param id id of the operation to query
     * @param clazz Class of the OperationResult subclass to construct.
     * @return Extended operation result object
     */
    protected <T extends OperationResult> T operationResult(String id, Class<T> clazz) {
        Exception ex = null;
        T result = null;
        Result baseResult = (this.resultDelegate != null) ? this.resultDelegate.getResult(this.engine, id) : this.engine.operationResult(id);
        try {
            result = (T) clazz.newInstance();

            // Pick off all the set* methods
            for (Method setter : clazz.getMethods()) {
                if (setter.getName().matches("set[A-Z][a-zA-Z0-9]*") && (setter.getParameterTypes().length == 1)) {
                    String field = ClassUtils.extractFieldNameFromBeanMethod(setter);

                    Method getter = ClassUtils.findGetterMethod(Result.class, field);
                    Object obj = null;
                    // First, see if the Result class from the workflow core has our field as
                    // a first-class citizen
                    if (getter != null) {
                        logger.internal("Populating result %s via %s", setter, getter);
                        obj = getter.invoke(baseResult);

                    }
                    // If not, and we're successful, see if it's in our output hashmap
                    else if (baseResult.getState().equals(OperationState.COMPLETE)) {
                        logger.internal("Populating output %s", field);
                        if (baseResult.getOutput() != null && baseResult.getOutput().containsKey(field)) {
                            obj = baseResult.getOutput().get(field);
                        }
                        else {
                            logger.error(MsgKey.FLOW_SOAP_RESULT_OUTPUT_NOT_DEFINED, baseResult.getType(), field);
                        }
                    }
                    setter.invoke(result, obj);
                }
            }
        }
        catch (InstantiationException e) {
            ex = e;
        }
        catch (IllegalAccessException e) {
            ex = e;
        }
        catch (IllegalArgumentException e) {
            ex = e;
        }
        catch (InvocationTargetException e) {
            ex = e;
        }
        if (ex != null) {
            throw new RuntimeException(logger.error(MsgKey.FLOW_SOAP_RESULT_REFLECTION_ERROR, ex, clazz.getName(), ex
                .getLocalizedMessage()), ex);
        }

        updateTag(result);

        return result;
    }
    
    /**
     * Find the most recent time activity of the result object. The last check point time
     * and the time of all messages are compared against each other. The update tag is set
     * to the most recent of all of these, if any are set.
     * 
     * @param result
     */
    protected void updateTag(OperationResult result) {
        Date oldestUpdate = result.getLastActiveTime();

        if (null != result.getMessages()) {
            for (LogMessage message : result.getMessages()) {
                if (null != message.getTimestamp()) {
                    if (null == oldestUpdate) {
                        oldestUpdate = message.getTimestamp();
                    }
                    else if (message.getTimestamp().after(oldestUpdate)) {
                        oldestUpdate = message.getTimestamp();
                    }
                }
            }
        }

        if (null != oldestUpdate) {
            result.setLastUpdate(new OperationUpdateTag(oldestUpdate.getTime()));
        }
    }
    /**
     * Loop through the LogMessages. Remove any log message that is older than the last update time
     * supplied by the caller.
     * 
     * @param operationResult
     * @param lastUpdateTag
     * @return Clone of operationResult with only messages newer than lastUpdateTag
     */
    protected OperationResult trimOldMessages(OperationResult operationResult, OperationUpdateTag lastUpdateTag) {
        OperationResult result = (OperationResult) operationResult.clone();
        
        if (null != lastUpdateTag) {
            Date lastUpdate = new Date(lastUpdateTag.getTag());
            
            if (null != operationResult.getMessages()) {
                // clearing out existing on clone so we don't have to use an explicit Iterator
                result.setMessages(new ArrayList<LogMessage>());
                
                for (LogMessage message : operationResult.getMessages()) {
                    if (message.getTimestamp().after(lastUpdate)) {
                        result.getMessages().add(message);
                    }
                }
            }
        }

        return result;
    }
}
