/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/OperatingSystem.java#1 $
 * 
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform;

import static com.netapp.common.platform.PlatformConfigBundle.DISABLE_CHANNEL_COPY;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import java.nio.channels.FileChannel;
import java.util.Collection;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.MessageKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;

/**
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public abstract class OperatingSystem {
    
    private static NALogger logger = NALogger.getLogger();
    
    private static final long MAX_BYTES_TO_COPY_PER_TRANSFER_CALL = 1024 * 1024 * 512;  // 500 MB
    
    protected String platform;
    protected ConfigBundle platformConfigBundle;
    private ProcessNative processNative;
    
    /*
     * INSTANTIATION METHODS
     */
    
    /**
     * Create a new instance according to the current OS
     *
     * @return OperatingSystem instance for the current OS
     */
    public static OperatingSystem createInstance() {
        return createInstance(OperatingSystem.determinePlatform());
    }

    /**
     * Create a new instance of the supplied platform. This is mostly
     * used for testing. {link createInstance()} should normally be used.
     *
     * @return OperatingSystem instance for the supplied OS
     */
    public static OperatingSystem createInstance(String platform) {
        OperatingSystem os = null;

        Exception ex = null;
        try {
            String className = "com.netapp.common.platform." + platform + ".OperatingSystemImpl";
            /*
             * Can't log anything until the instance is created (since the logger depends upon us,
             * this would create a circular dependency) 
             */
            // logger.internal("Looking for OperatingSystem handler " + className);

            Class<?> clazz = Class.forName(className);
            os = (OperatingSystem)clazz.getConstructor().newInstance();
            os.platform = platform;
            // logger.debug(PLATFORM_OPERATING_SYSTEM, new Object[] { os.getPlatform(), os.getVersion(), os.getArch() });
        }
        catch (ClassNotFoundException e) {
            ex = e;
        }
        catch (NoSuchMethodException e) {
            ex = e;
        }
        catch (IllegalAccessException e) {
            ex = e;
        }
        catch (IllegalArgumentException e) {
            ex = e;
        }
        catch (InstantiationException e) {
            ex = e;
        }
        catch (InvocationTargetException e) {
            ex = e;
        }
        if (ex != null) {
            throw new PlatformException(MessageBundle.localizeMessage(
                MsgKey.PLATFORM_UNSUPPORTED, platform), ex);
        }
        return os;
    }
    

    /**
     * Return the normalized identifier for the current platform.
     * The result of System.getProperty("os.name") is used to lookup the
     * correct identifier in the global properties file. If that fails,
     * the lowercase value of os.name is returned.
     * This is only used by OperatingSystem to bootstrap and a couple
     * test hooks. Do not use this elswhere, use OperatingSystem.getPlatform
     * instead.
     * 
     * @return standardized platform identifier
     */
    protected static String determinePlatform() {
        String osName = System.getProperty("os.name");
        
        // We use a hardcoded config bundle here because we are shipping a default platform.config file with the NN jar file
        // and applications can supply their own to override if necessary.  We also can't construct this statically 
        // because of the circular dependency with NALogger.
        ConfigBundle platformConfigBundle = new PlatformConfigBundle();
        ConfigProperty<String> platformProperty = new ConfigProperty<String>(String.class, "os." + osName);
        String platform = platformProperty.lookup(platformConfigBundle);
        if (platform == null || platform.length() == 0) {
            platform = osName.toLowerCase();
        }
        return platform;
    }
    
    /**
     * Return the string representing the current platform instance.
     * This is always the most specific package enclosing the concrete
     * OperatingSystem class.
     *
     * @return A representation of the platform this instance supports
     */
    public String getPlatform() {
        return this.platform;
    }
    
    /**
     * Return a human-readable string representing the current platform instance.
     * Subclasses may override this if desired.
     * 
     * @return A non-canonical representation of the platform for this instance
     */
    public String getPlatformBrand() {
        return System.getProperty("os.name");
    }
    /**
     * Return the "basic" platform String - either "unix" or "windows".
     * 
     * Useful for forming composite keys based on this String (e.g. for
     * indexing into a properties file).
     * 
     * For determining whether the platform is Unix or Windows, use 
     * {@link #isUnix()} or {@link #isWindows()} instead.  
     */
    public String getBasicPlatform() {
        return this.isWindows() ? "windows" : "unix";
    }

    /**
     * Return true only if the operating system is Windows.
     *
     * @return true only if the operating system is Windows.
     */
    public boolean isWindows() {
        return false;
    }

    public boolean isUnix() {
        return !this.isWindows();
    }
    
    /**
     * Return the version of the current platform. String follows the
     * convention of local platform where applicable. By default, this
     * is the same as System.getProperty("os.version") but on some platforms
     * may differ.
     * Examples:
     * getPlatform() = "aix": "5.2", "5.3" 
     * getPlatform() = "solaris": 5.7, 5.8, 5.9
     * getPlatform() = "linux": "RHEL 3", "RHEL 4", "SEL 9"
     *
     * @return Version of the current platform instance
     */
    public String getVersion() {
        return System.getProperty("os.version");
    }

    /**
     * Return a representation of the current architecture. 
     * By default, this is the same as System.getProperty("os.arch")
     *
     * @return Architecture of the current platform
     */
    public String getArch() {
        return System.getProperty("os.arch");
    }

    /*
     * ABSTRACT METHODS
     */
    
    /**
     * Get the PID of the current process.
     * @return the PID of the current process
     */
    public abstract Long getPID();
    
    /**
     * Kill a process
     * @param pid PID of the process
     * @param signal Signal to send (e.g. 9)
     */
    public abstract void kill(long pid, int signal);
    
    public abstract boolean isProcessRunning(Long pid);
    
    public abstract boolean isUserPrivileged();
    
    /**
     * Get the name of the environment variable that controls the
     * dynamic linker path.
     *
     * @return The name of the environment variable to control the dynamic linker.
     */
    public abstract String getLibraryEnvironmentName();
    
    public abstract DeviceAnalyzer createDeviceAnalyzer();

    public abstract File getUserApplicationDirectory();

    /**
     * Lookup the User object of the specified user name
     * @param username Username to lookup
     * @return User of Username
     * @throws CredentialNotFoundException if the username is not found
     */
    public abstract User getUserFromUsername(String username);
    
    /**
     * Lookup the Group object of the specified group name
     * @param groupname Username to lookup
     * @return Group of group name
     * @throws CredentialNotFoundException if the groupname is not found
     */
    public abstract Group getGroupFromGroupname(String groupname);

    /**
     * Change ownership and permissions on a file or directory.
     * 
     * @param file File or directory to be changed
     * @param securityDescriptor The new security descriptor settings to use
     */
    public abstract void changeFileSecurity(File file, FileSecurityDescriptor securityDescriptor);
    
    /**
     * Change the owner (user and group) on a file or directory.
     * 
     * @see #getUserFromUsername(String)
     * @see #getGroupFromGroupname(String)
     * 
     * @param path
     * @param fileOwner The new user and group that should own the file or directory.
     */
    public abstract void changeOwner(File path, FileOwner fileOwner);
    
    /**
     * Creates a new instance of a file security descriptor with the parameters filled out.
     * 
     * @param user The user that is the file owner
     * @param group The group that is the file owner
     * @param accessControls The set of permissions that should be granted.
     * @return A newly created security descriptor that can be used with changeFileSecurity
     */
    public abstract FileSecurityDescriptor createFileSecurityDescriptor(User user, Group group, FileSecurityDescriptor.AccessControl...accessControls);
    
    /**
     * Return the list of currently mounted filesystems.
     * 
     * @return list of currently mounted filesystems.
     */
    public abstract Collection<FileSystemMount> getMountedFilesystems();
    
    /**
     * Returns the UUID for the current operating system.
     * 
     * @return The current UUID as a string
     */
    public abstract String getBiosUuid(); 
    
    /*
     * CONCRETE METHODS
     */

    public final File getUserHomeDir() {
        return new File(System.getProperty("user.home"));
    }
    
    public final String getUsername() {
        return System.getProperty("user.name");
    }

    /**
     * Lookup the User object for the current user that is executing 
     * this java process
     */
    public final User getUser() {
        return this.getUserFromUsername(this.getUsername());
    }
    
    /**
     * Change the owner on a file or directory and any directories below.
     * 
     * @param root
     * @param fileOwner TODO
     */
    public void changeOwnerRecursive(File root, FileOwner fileOwner) {
        this.changeOwner(root, fileOwner);
        if (root.isDirectory()) {
            File[] contents = root.listFiles();
            for (File f : contents) {
                this.changeOwnerRecursive(f, fileOwner);
            }
        }
    }

    /**
     * Copy a SMOFile from source to destination
     * Subclasses can override to perform more efficient copying.
     * 
     * @param src Source file.  Must exist and be readable.
     * @param dest Destination for copy.  If dest is a file, it must not exist but its directory must be writable.  If dest is a directory, it must exist and be writable. 
     */
    public void copyFile(File src, File dest) throws IOException {
        try {
            logger.info(MsgKey.PLATFORM_COPY_FILE_STARTING, src, dest);
            
            // The src.length() != 0 check is in here for copying raw devices.
            // Sometimes raw devices report file length == 0.
            // copyFileUsingChannels needs the file length to copy.
      
            if (!DISABLE_CHANNEL_COPY.lookup(this.getPlatformConfigBundle()) && src.length() != 0) {
                try {
                    copyFileUsingChannels(src, dest);
                }
                catch (IOException e) {
                    logger.internal("Could not copy using channels", e);
                    // ignore the exception using channels, retry using Streams
                    copyFileUsingStreams(src, dest);
                }
            }
            else {
                copyFileUsingStreams(src, dest);
            }
        }
        catch (IOException e) {
            logger.debug(MsgKey.PLATFORM_COPY_FILE_FAILED, src, dest, e.getLocalizedMessage());            
            throw e;
        } 
    }

    /**
     * Copies files by using {@link FileChannel#transferTo(long, long, java.nio.channels.WritableByteChannel)}.
     * 
     * Do not use this method for files whose file.length() returns 0 (ie some raw devices).
     * 
     * This method needs a file length in order to work properly.
     * 
     * @param src
     * @param dest
     * @throws IOException
     */
    protected void copyFileUsingChannels(File src, File dest) throws IOException {
        
        assert src.length() != 0;
        
        FileInputStream inStream = null;
        FileOutputStream outStream = null;
        try {
            // Open the input file for reading.
            // Open the output file for writing. If it does not exist, it will
            // be created. If it exists, it will be overwritten.
            logger.internal("Attempting to copy from " + src + " to " + dest + " using channels");
      
            inStream = new FileInputStream(src);
            outStream = new FileOutputStream(dest);
            
            FileChannel inChannel = inStream.getChannel();
            FileChannel outChannel = outStream.getChannel();
            
            long size = inChannel.size();
            long remainingBytes = size;
            while (remainingBytes > 0) {
                // Received "java.io.IOException: Cannot allocate memory" when trying to transfer a 200 GB file,
                // so break it up into 500 MB chunks                
                remainingBytes -= inChannel.transferTo(size - remainingBytes, Math.min(remainingBytes, MAX_BYTES_TO_COPY_PER_TRANSFER_CALL), outChannel);
            }
        }
        finally {
            if (inStream != null) {
                inStream.close();
            }
            if (outStream != null) {
                outStream.close();    
            }
        }
    }

    /**
     * Copies files using {@link FileInputStream} and {@link FileOutputStream}
     * by reading bytes from the inputstream into a byte array buffer
     * and then writing them to the output stream.
     * 
     * @param src
     * @param dest
     * @throws IOException
     */
    protected void copyFileUsingStreams(File src, File dest) throws IOException {
        InputStream in = null;
        OutputStream out = null;
        try {
      
            logger.internal("Attempting to copy from " + src + " to " + dest + " using streams");
            in = new BufferedInputStream(new FileInputStream(src));
            out = new BufferedOutputStream(new FileOutputStream(dest));
      
            byte[] buf = new byte[4096];
            int n;
            while ((n = in.read(buf)) != -1) {
                out.write(buf, 0, n);
            }
        }
        finally {
            if (in != null) {
                in.close();
            }
            if (out != null) {
                out.close();    
            }
        }
    }

    public abstract MountpointMapper createMountpointMapper();

    public abstract FileSecurityDescriptor getFileSecurityDescriptor(File file);
    
    public boolean canExecute(File executeable, User user) {
        return false;
    }
    
    public abstract PartitionTable readPartitionTable(File device);
    
    protected synchronized ProcessNative getProcessNative() {
        if (this.processNative == null) {
            try {
                Class<?> processNativeClass = Class.forName("com.netapp.common.platform.ProcessNativeImpl");
                this.processNative = (ProcessNative)processNativeClass.newInstance();
            }
            catch (UnsatisfiedLinkError e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (ClassNotFoundException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (InstantiationException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
            catch (IllegalAccessException e) {
                logger.warn(MsgKey.PLATFORM_PROCESS_UNABLE_TO_LOAD_NATIVE, e, e.getLocalizedMessage());
            }
        }
        
        if (this.processNative == null) {
            throw new UnsupportedOperationException(logger.error(MsgKey.PLATFORM_PROCESS_METHOD_DEPENDS_ON_NATIVE_LIBRARY));
        }
        
        return this.processNative;
    }

    public synchronized ConfigBundle getPlatformConfigBundle() {
        if (this.platformConfigBundle == null) {
            this.platformConfigBundle = new PlatformConfigBundle();
        }
        return this.platformConfigBundle;
    }
    
    /**
     * Returns the canonical path of the file, but
     * if that fails, returns the absolute path of the file.
     *  
     * @param file File for which to get the path
     * @return the canonical path of the file, but
     * if that fails, returns the absolute path of the file.
     */
    protected String getFilePath(File file) {
        String path;
        try {
            path = file.getCanonicalPath();
        }
        catch (IOException e) {
            // ignore, just use absolute path
            path = file.getAbsolutePath(); 
        }
        return path;
    }

    /**
     * Returns true if the specified user is a member of the group
     * 
     * @param username The user, either specified by name or id, to be checked
     * @param groupname The group, either specified by name or id, of the file's attribute  
     * @return 
     * <ul>
     *   <li>true, if the user is a member of the group</li>
     *   <li>false, if the user is not a member of the group</li>
     * </ul>
     */
    public abstract boolean userInGroup(String username, String groupname);
    
    /**
     * Creates and returns an Authenticator appropriate for the current OS
     * @param configBundle The value 'pam.service.name' will be used to uniquely identify the authentication requests and a file should exist in /etc/pam.d of the same name.
     * @return An appropriate Authenticator for the current OS
     */
    public abstract Authenticator createAuthenticator(ConfigBundle configBundle);
    
    /**
     * Log a message to the Operating System logger, if available. Otherwise
     * this is a no-op. 
     * @param level LogLevel for this message. This may be interpeted in a system-dependent manner
     * @param source Application name to use when required
     * @param key MsgKey containing the message to log
     */
    public void systemEvent(LogLevel level, String source, MessageKey key) {
    }

    /**
     * Log a message to the Operating System logger, if available. Otherwise
     * this is a no-op. 
     * @param level LogLevel for this message. This may be interpeted in a system-dependent manner
     * @param source Application name to use when required
     * @param key MsgKey containing the message to log
     * @param t Exception associated with this message. This may be used in a system-dependent manner
     */
    public void systemEvent(LogLevel level, String source, MessageKey key, Throwable t) {
    }

    /**
     * Log a message to the Operating System logger, if available. Otherwise
     * this is a no-op. 
     * @param level LogLevel for this message. This may be interpeted in a system-dependent manner
     * @param source Application name to use when required
     * @param key MsgKey containing the message to log
     * @param params Parameters for localizing key
     */
    public void systemEvent(LogLevel level, String source, MessageKey key, Object... params) {
    }

    /**
     * Log a message to the Operating System logger, if available. Otherwise
     * this is a no-op. 
     * @param level LogLevel for this message. This may be interpeted in a system-dependent manner
     * @param source Application name to use when required
     * @param key MsgKey containing the message to log
     * @param params Parameters for localizing key
     * @param t Exception associated with this message. This may be used in a system-dependent manner
     */
    public void systemEvent(LogLevel level, String source, MessageKey key, Throwable t, Object... params) {
    }
    
    @Override
    public String toString() {
        return MessageBundle.localizeMessage(MsgKey.PLATFORM_DEFAULT_DESCRIPTION,
            getPlatformBrand(), getVersion(), getArch());
    }
    
    public static void main(String[] args) {
        OperatingSystem os = OperatingSystem.createInstance();
        if (args.length == 1 && "-libpath".equals(args[0])) {
            // Print the canonical library "path"
            System.out.println(os.getPlatform() + File.separator + os.getArch());
        }
        else {
            // Dump our full information
            System.out.println("Running " + os);
            System.out.println("Platform: " + os.getPlatformBrand() + " (" + os.getPlatform() + ")");
            System.out.println("Version: " + os.getVersion());
            System.out.println("Arch: " + os.getArch());
        }
    }
}
