/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/linux/RawDeviceAnalyzer.java#1 $
 * 
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.linux;

import java.io.File;
import java.util.Collection;
import java.util.Collections;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.MsgKey;
import com.netapp.common.execution.ShellCommand;
import com.netapp.common.execution.ShellResult;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.DeviceType;
import com.netapp.common.platform.ManagedRawDeviceMapper;
import com.netapp.common.platform.OperatingSystem;
import com.netapp.common.platform.PlatformException;
import com.netapp.common.platform.RawDeviceMapping;
import com.netapp.common.util.ConfigProperty;

/**
 * DeviceAnalyzer that manages /dev/raw devices on 2.4 and older 2.6 distributions.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class RawDeviceAnalyzer extends LinuxDeviceAnalyzer implements ManagedRawDeviceMapper {
    private static final NALogger logger = NALogger.getLogger(RawDeviceAnalyzer.class);
    
    /**
     * raw device mapping timeout value in milliseconds - timeout for raw -q to succeed
     */
    protected static final ConfigProperty<Long> RAW_DEVICE_MAPPING_TIMEOUT_MS = ConfigProperty.longProperty("rawdevice.mapping.timeoutMs", 20000L);
    
    /**
     * raw device unmapping delay to avoid 'device busy' error in suse9     
     */
    protected static final ConfigProperty<Long> RAW_DEVICE_BEFORE_UNMAPPING_FIXED_WAIT_MS = ConfigProperty.longProperty("rawdevice.before.unmapping.waitMs", 5000L);
    
    /**
     * raw device mapping delay to after 'raw -q' succeeds to avoid chown failing in suse9     
     */
    protected static final ConfigProperty<Long> RAW_DEVICE_MAPPING_FIXED_WAIT_MS = ConfigProperty.longProperty("rawdevice.mapping.fixed.waitMs", 5000L);
       
    /**
     * Matches "/dev/raw/raw51: bound to major 8, minor 177" 
     */
    private static final Pattern RAW_QUERY_RESULT_PATTERN = Pattern.compile(".*?bound to major (\\d+), minor (\\d+)");
    protected static final Pattern RAW_DEVICE_PATTERN = Pattern.compile("/dev/raw/raw(\\d+).*");
    /**
     * The max raw device number we expect to find
     */
    protected static final int MAX_RAW_DEVICE_NUMBER = 255;
    
    
    public RawDeviceAnalyzer(OperatingSystem operatingSystem) {
        super(operatingSystem);
    }
    
    public DeviceType determineDeviceType(File file) {
        if (RAW_DEVICE_PATTERN.matcher(file.getAbsolutePath()).matches()) {
            return DeviceType.RAW_DEVICE;
        }
        return super.determineDeviceType(file);
    }

    /**
     * 
     * @see com.netapp.common.platform.ManagedRawDeviceMapper#createAndAddMapping(boolean, java.io.File)
     */
    public RawDeviceMapping createAndAddMapping(boolean isPersistent, File disk) {
        return this.createAndAddMapping(isPersistent, disk, this.autoGenerateRawDeviceName());
    }

    /**
     * 
     * @see com.netapp.common.platform.ManagedRawDeviceMapper#createAndAddMapping(boolean, java.io.File, java.io.File)
     */
    public RawDeviceMapping createAndAddMapping(boolean isPersistent, File disk, File rawDevice) {
        logger.info(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_CREATING_RAW_DEVICE_MAPPING_LINUX, new Object[] { disk, rawDevice });
        RawDeviceMapping rawDeviceMapping = new RawDeviceMapping(rawDevice, disk);
        getShellExecutor().execute(new ShellCommand("raw", rawDevice.getAbsolutePath(), disk.getAbsolutePath()));
        
        waitForDeviceMappingDone(rawDevice);
        
        if (isPersistent) {
            RawDevicesFile rawDevicesFile = this.getRawDevicesFile();
            rawDevicesFile.addMapping(rawDeviceMapping);
        }
        
        return rawDeviceMapping;
    }

    /**
     * 
     * @see com.netapp.common.platform.ManagedRawDeviceMapper#removeMapping(com.netapp.common.platform.RawDeviceMapping)
     */
    public void removeMapping(RawDeviceMapping rawDeviceMapping) {
        File rawDevice = rawDeviceMapping.getRawDevice();
        File disk = rawDeviceMapping.getDisk();
        logger.info(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_REMOVING_RAW_DEVICE_MAPPING_LINUX, new Object[] { disk, rawDevice });
        
        Long rawDevUnmappingFixedWaitInMillis = 0L;
        if ((this.operatingSystem == null) || (!(this.operatingSystem instanceof OperatingSystemImpl)))
        {
            rawDevUnmappingFixedWaitInMillis = RAW_DEVICE_BEFORE_UNMAPPING_FIXED_WAIT_MS.getDefaultValue();
        }
        else {
            rawDevUnmappingFixedWaitInMillis = RAW_DEVICE_BEFORE_UNMAPPING_FIXED_WAIT_MS.lookup(((OperatingSystemImpl)this.operatingSystem).getPlatformConfigBundle());
        }
        logger.internal("raw device unmapping fixed wait  in ms = " + rawDevUnmappingFixedWaitInMillis);
        
        /*
         * Adding an unconditional sleep in this function as we are seeing intermittent failures
         * in unmapping the raw device file saying 'device or resource busy' and fuser command
         * shows oracle is still using the device. With a delay, it works.
         * This issue is found on SUSE9 only as of now. 
         */
        logger.internal("sleeping at the start of removeMapping for raw device " + rawDevice.getAbsolutePath());
        
        try {
            Thread.sleep(rawDevUnmappingFixedWaitInMillis);            
        }
        catch (InterruptedException e) {
            logger.internal("Caught exception " , e);
            // Restore the interrupted status
            Thread.currentThread().interrupt();
        }
        
        logger.internal("coming out of sleep in removeMapping for raw device" + rawDevice.getAbsolutePath());
        
        // Unbind the raw device by setting its major/minor to 0, 0
        // http://mail.nl.linux.org/linux-mm/2000-12/msg00012.html
        ShellResult shellResult = getShellExecutor().execute(new ShellCommand("raw", rawDevice.getAbsolutePath(), "0", "0"));
        if (!shellResult.getStdErr().isEmpty()) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_UNMAP_RAW_DEVICE, rawDevice, shellResult);
            throw new PlatformException(error);
        }

        RawDevicesFile rawDevicesFile = this.getRawDevicesFile();
        rawDevicesFile.removeMapping(rawDeviceMapping);
    }

    /**
     * 
     * @see com.netapp.common.platform.DeviceAnalyzer#lookupMappingByDisk(java.io.File)
     */
    public RawDeviceMapping lookupMappingByDisk(File disk) {
        String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNSUPPORTED_OPERATION);
        throw new PlatformException(error);
    }

    /**
     * 
     * @see com.netapp.common.platform.DeviceAnalyzer#lookupMappingByRawDevice(java.io.File)
     */
    public RawDeviceMapping lookupMappingByRawDevice(File rawDevice) {
        ShellResult rawQueryResult = getShellExecutor().execute(new ShellCommand("raw", "-q", rawDevice.getAbsolutePath()));
        if (!rawQueryResult.getStdErr().isEmpty() || rawQueryResult.getStdOut().size() != 1) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_MAP_RAW_DEVICE, rawDevice, rawQueryResult.getAllOut());
            throw new PlatformException(error);    
        }
        
        String queryOutput = rawQueryResult.getStdOut().get(0);
        Matcher queryMatcher = RAW_QUERY_RESULT_PATTERN.matcher(queryOutput);
        
        if (!queryMatcher.matches()) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_MAP_RAW_DEVICE, rawDevice, rawQueryResult.getAllOut());
            throw new PlatformException(error);
        }
        
        int majorNumber = Integer.parseInt(queryMatcher.group(1));
        int minorNumber = Integer.parseInt(queryMatcher.group(2));

        if (majorNumber == 0 && minorNumber == 0) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_MAP_RAW_DEVICE, rawDevice, rawQueryResult.getAllOut());
            throw new PlatformException(error);
        }
        
        File disk = this.deviceFromMajorMinor(majorNumber, minorNumber);
        if (disk == null) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_MAP_RAW_DEVICE, rawDevice, rawQueryResult.getAllOut());
            throw new PlatformException(error);
        }
        return new RawDeviceMapping(rawDevice, disk);
    }
    
    @SuppressWarnings("unchecked")
    public File autoGenerateRawDeviceName() {
        return autoGenerateRawDeviceName((Collection<String>) Collections.EMPTY_SET);
    }

    public File autoGenerateRawDeviceName(Collection<String> unusableNames) {
        boolean[] availableRawDeviceNames = new boolean[256];
        for (int i = 1; i <= 255; i++) {
            availableRawDeviceNames[i] = true;
        }

        // first, we use 'raw -qa' to find all current mappings in use
        ShellResult rawQueryResult = getShellExecutor().execute(new ShellCommand("raw", "-qa"));
        for (int i = 0; i < rawQueryResult.getStdOut().size(); i++) {
            String queryOutput = rawQueryResult.getStdOut().get(i);
            Matcher queryMatcher = RAW_DEVICE_PATTERN.matcher(queryOutput);
            if (!queryMatcher.matches()) {
                // this one should be ok, not a recognized output line
                continue;
            }
            int deviceNumber = Integer.parseInt(queryMatcher.group(1));

            queryMatcher = RAW_QUERY_RESULT_PATTERN.matcher(queryOutput);
            if (!queryMatcher.matches()) {
                // this is bad, we found a device entry but can't determine what it is bound to
                String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_AUTO_GENERATE_RAW_DEVICE_NAME, queryOutput);
                throw new PlatformException(error);
            }

            int majorNumber = Integer.parseInt(queryMatcher.group(1));
            int minorNumber = Integer.parseInt(queryMatcher.group(2));
            logger.internal("Found raw device " + deviceNumber + " with majorNumber " + majorNumber
                + " and minorNumber " + minorNumber);
            if (majorNumber == 0 && minorNumber == 0) {
                availableRawDeviceNames[deviceNumber] = true;
            }
            else {
                availableRawDeviceNames[deviceNumber] = false;
            }
        }

        /*
         * Now, we parse the raw devices file for any persistent rawdevices
         */
        RawDevicesFile rawDevicesFile = this.getRawDevicesFile();
        for (RawDeviceMapping rawDeviceMapping : rawDevicesFile.getRawDeviceMappings()) {
            Matcher matcher = RAW_DEVICE_PATTERN.matcher(rawDeviceMapping.getRawDevice().getAbsolutePath());
            if (matcher.matches()) {
                int deviceNumber = Integer.parseInt(matcher.group(1));
                logger.internal("Found device " + deviceNumber + " in " + rawDevicesFile.getFile());
                if (deviceNumber <= MAX_RAW_DEVICE_NUMBER) {
                    availableRawDeviceNames[deviceNumber] = false;
                }                
            }
        }
    
        // finally, return the first match
        for (int i = 1; i <= 255; i++) {
            if (availableRawDeviceNames[i]) {
                String candidateName = "/dev/raw/raw" + i;
                if (!unusableNames.contains(candidateName)) {
                    logger.internal("Auto-generated raw device candidate " + candidateName);
                    return new File(candidateName);
                }
            }
        }

        String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_AUTO_GENERATE_RAW_DEVICE_NAME, "");
        throw new PlatformException(error);
    }
    
    protected RawDevicesFile getRawDevicesFile() {
        if (this.operatingSystem instanceof OperatingSystemImpl) {
            return ((OperatingSystemImpl)this.operatingSystem).getRawDevicesFile();
        }
        throw new RuntimeException(
            logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
    }
    
    private boolean isDeviceMappingDone(File rawDevice) {
        ShellResult rawQueryResult = getShellExecutor().execute(new ShellCommand("raw", "-q", rawDevice.getAbsolutePath()));
        if (!rawQueryResult.getStdErr().isEmpty() || rawQueryResult.getStdOut().size() != 1) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_MAP_RAW_DEVICE, rawDevice, rawQueryResult.getAllOut());
            throw new PlatformException(error);    
        }
        
        String queryOutput = rawQueryResult.getStdOut().get(0);
        Matcher queryMatcher = RAW_QUERY_RESULT_PATTERN.matcher(queryOutput);
        
        if (!queryMatcher.matches()) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNABLE_TO_MAP_RAW_DEVICE, rawDevice, rawQueryResult.getAllOut());
            throw new PlatformException(error);
        }
        
        int majorNumber = Integer.parseInt(queryMatcher.group(1));
        int minorNumber = Integer.parseInt(queryMatcher.group(2));

        if (majorNumber == 0 && minorNumber == 0) {
            logger.debug(MsgKey.PLATFORM_LINUX_RAWDEVICEANALYZER_RAW_DEVICE_MAPPING_NOT_DONE, rawDevice, rawQueryResult.getAllOut());
            return false;            
        }
        
        logger.debug(MsgKey.PLATFORM_LINUX_RAWDEVICEANALYZER_RAW_DEVICE_MAPPING_DONE, rawDevice, rawQueryResult.getAllOut());
        return true;                
    }
    
    private boolean waitForDeviceMappingDone(File rawDevice) {
        Long mappingTimeoutInMillis = 0L;
        Long mappingFixedWaitMillis = 0L;
        
        if ((this.operatingSystem == null) || (!(this.operatingSystem instanceof OperatingSystemImpl)))
        {
            mappingTimeoutInMillis = RAW_DEVICE_MAPPING_TIMEOUT_MS.getDefaultValue();
        }
        else {
            mappingTimeoutInMillis = RAW_DEVICE_MAPPING_TIMEOUT_MS.lookup(((OperatingSystemImpl)this.operatingSystem).getPlatformConfigBundle());
        }
        
        logger.internal("raw device mapping timeout in ms = " + mappingTimeoutInMillis);
        
        long startTime = 0;
        while (!isDeviceMappingDone(rawDevice) && (startTime < mappingTimeoutInMillis)) {
            try {
                Thread.sleep(100L);
                startTime += 100;                 
                logger.debug(MsgKey.PLATFORM_LINUX_RAWDEVICEANALYZER_WAITING_FOR_RAW_DEVICE_MAPPING, rawDevice.getAbsolutePath());
            }
            catch (InterruptedException e) {
                logger.internal("Caught exception " , e);
                // Restore the interrupted status
                Thread.currentThread().interrupt();
                break;
            }
        }
        
        if ((this.operatingSystem == null) || (!(this.operatingSystem instanceof OperatingSystemImpl)))
        {
            mappingFixedWaitMillis = RAW_DEVICE_MAPPING_FIXED_WAIT_MS.getDefaultValue();
        }
        else {
            mappingFixedWaitMillis = RAW_DEVICE_MAPPING_FIXED_WAIT_MS.lookup(((OperatingSystemImpl)this.operatingSystem).getPlatformConfigBundle());
        }
        
        logger.internal("raw device mapping fixed wait  in ms = " + mappingFixedWaitMillis);
        
        /*
         * Adding an unconditional sleep in this function as we are seeing intermittent failures
         * in chowning the raw device file even though raw -q has given valid output.
         * This issue is found on SUSE9 only as of now. 
         */
        logger.internal("sleeping at the end of waitForDeviceMappingDone for raw device" + rawDevice.getAbsolutePath());
        try {
            Thread.sleep(mappingFixedWaitMillis);            
        }
        catch (InterruptedException e) {
            logger.internal("Caught exception " , e);
            // Restore the interrupted status
            Thread.currentThread().interrupt();
        }

        logger.internal("coming out of sleep at the end of waitForDeviceMappingDone for raw device" + rawDevice.getAbsolutePath());
        
        return (isDeviceMappingDone(rawDevice));
    }
}
