/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/linux/smbios/SmBiosParser.java#1 $
 * 
 * Copyright (c) 2009 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.linux.smbios;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

import com.netapp.common.logging.NALogger;

/**
 * SM BIOS is a specification created by the DMTF which can be found at
 * http://www.dmtf.org/standards/smbios/. This is a 'clean room' implementation
 * of the required parts of the spec in order to retrieve elements that we care
 * about. For now, this is just the UUID of the operating system.
 * 
 * The spec tells us to use the following algorithm:
 * 
 * Access memory starting at address 0x000F0000. Read in 16 byte paragraphs and
 * look for the first paragraph that starts with '_SM_'. This paragraph must be
 * found before address 0X000FFFFF.
 * 
 * When the SM BIOS paragraph is found, use the BasicStructuredHeader to find
 * out what the physical RAM address is for the first defined structure.
 * 
 * Each of these structures have a standard header section that defines the type
 * of structure and a length. The length is only partly true. Even if we skip to
 * length number of bytes forward, we still need to continue reading until the
 * next double-null is reached. That's the true length of the structure,
 * including text definitions.
 * 
 * We must read these structures looking for the SystemInformationHeader type.
 * Once found, we can then read out the elements of the UUID and the create the
 * UUID string that is to be sent back to the user.
 * 
 * @author josh
 * 
 */
public class SmBiosParser {
    private static final NALogger LOGGER                    = NALogger.getLogger(SmBiosParser.class);

    private static final String   LINUX_MEMORY_DEVICE       = "/dev/mem";
    private static final String   FILE_READ_ACCESS          = "r";
    private static final String   SM_BIOS_IDENTIFIER_STRING = "_SM_";

    private static final long     BIOS_START_ADDRESS        = 0x000F0000;
    private static final long     BIOS_END_ADDRESS          = 0x000FFFFF;
    private static final int      BIOS_PARAGRAPH_SIZE       = 16;

    private RandomAccessFile      memoryFile;

    protected SmBiosParser() {
        try {
            RandomAccessFile file = new RandomAccessFile(LINUX_MEMORY_DEVICE, FILE_READ_ACCESS);
            setMemoryFile(file);
        }
        catch (FileNotFoundException e) {
            throw new SecurityException(e);
        }
    }

    /**
     * Read the OS UUID from memory following the SM BIOS specification.
     * 
     * @throws SecurityException
     *             thrown if the user does not have permission to access
     *             /dev/mem.
     * @throws IOException
     *             thrown if there is a problem reading particular bytes from
     *             /dev/mem
     * @return
     */
    public static String readUuid() throws IOException {
        SmBiosParser parser = new SmBiosParser();

        parser.seekToTableOffset();
        parser.findSystemInformationTable();
        SystemInformationHeader header = SystemInformationHeader.read(parser.getMemoryFile());

        return header.getUuid();
    }

    /**
     * To find the system information table, we have to start reading in the
     * table header structure looking for a particular type that signifies the
     * system information table. End looking if we get the 'end of tables'
     * table.
     * 
     * This method really should use recursion, but I'm afraid the recursion
     * will go too deep
     * 
     * @throws IOException
     */
    protected void findSystemInformationTable() throws IOException {
        BasicStructuredHeader header = BasicStructuredHeader.read(getMemoryFile());
        LOGGER.internal("Found table type %d, handle %d, length %d", header.getType(), header.getHandle(), header.getLength());

        if (BasicStructuredHeader.SYSTEM_INFORMATION_TYPE != header.getType()) {
            // skip to the end of the structured data. Do not forget that the
            // length must
            // account for already reading the basic header.
            byte [] toSkip = new byte[header.getLength() - BasicStructuredHeader.HEADER_IN_BYTES];
            getMemoryFile().read(toSkip);

            seekTableEnd();
            findSystemInformationTable();
        }
    }

    /**
     * In order to seek to the end of a table, we have to read bytes until we
     * find two consecutive null values.
     * 
     * @throws IOException
     */
    protected void seekTableEnd() throws IOException {
        boolean isNull = false;

        while (true) {
            int current = getMemoryFile().read();

            if (0 == current) {
                if (isNull) {
                    // at end of table
                    return;
                }

                isNull = true;
            }
            else {
                isNull = false;
            }
        }
    }

    /**
     * The SM BIOS table header has to be searched for. The search algorithm is
     * defined as:
     * 
     * <ul>
     * <li>Start at position 0x000F0000
     * <li>Read in 16-byte pages at a time
     * <li>If a page starts with '_SM_', then it is the start of the SM BIOS
     * table header
     * <li>End at position 0x000FFFFF
     * </ul>
     * 
     * @throws IOException
     * 
     */
    protected void seekToTableOffset() throws IOException {
        long offset = -1;

        // map memory from 0x000F0000(hex) - 0x000FFFFF(hex) so that we can
        // search through it
        long position = BIOS_START_ADDRESS;

        // Walk the buffer in 16-byte pages. Find the page that starts with
        // '_SM_'
        int pageCount = 0;

        do {
            getMemoryFile().seek(position);

            // read the first four bytes
            byte[] pageStart = new byte[4];
            getMemoryFile().read(pageStart);

            String anchor = new String(pageStart);

            if (anchor.equals(SM_BIOS_IDENTIFIER_STRING)) {
                // retrieve the structure table address value from offset
                // 0x18(hex). This is a DWORD (4 bytes) which is a Java int
                getMemoryFile().seek(position + 0x18);

                // read a 4-byte integer from this position
                offset = ByteOrderConverter.readInNativeOrder(getMemoryFile().readInt());
                LOGGER.internal("Found the SM BIOS structure table address: %s", offset);
                break;
            }

            position += BIOS_PARAGRAPH_SIZE;
            pageCount++;
        }
        while (position < BIOS_END_ADDRESS);

        LOGGER.internal("Processed %d pages", pageCount);
        LOGGER.internal("Seeking to offset %s for the first structured table", offset);
        getMemoryFile().seek(offset);
    }

    protected RandomAccessFile getMemoryFile() {
        return memoryFile;
    }

    protected void setMemoryFile(RandomAccessFile memoryFile) {
        this.memoryFile = memoryFile;
    }
    
    /**
     * Main method to enable manual testing of this class 
     * 
     * @param args
     */
    public static void main(String[] args) {
        try {
            System.out.println("Found UUID = " + SmBiosParser.readUuid());
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }

}
