/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/unix/UnixMountpointMapper.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.unix;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.netapp.common.platform.MountpointMapper;
import com.netapp.common.util.CollectionUtils;
import com.netapp.common.util.ConfigProperty;

/**
 * Map a set of files to their mountpoints on a Unix OS
 * 
 * <b>NOTE: Can return a mountpoint on non-NetApp storage</b>
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class UnixMountpointMapper implements MountpointMapper {

    private OperatingSystemImpl operatingSystem;
    
    public UnixMountpointMapper(OperatingSystemImpl operatingSystem) {
        this.operatingSystem = operatingSystem;
    }
    
    public Map<File, File> map(Set<File> fileSet) {
        Long numFiles = 3L;
        ConfigProperty<Long> filesPerDfCommand;
        if ((this.operatingSystem == null) || (!(this.operatingSystem instanceof OperatingSystemImpl)))
        {
            numFiles = 3L;
        } 
        else {
            if (this.operatingSystem.getPlatform().equalsIgnoreCase("aix")) {
                filesPerDfCommand = ConfigProperty.longProperty("aix.files.per.df.command", 3L);
            }
            else {
                filesPerDfCommand = ConfigProperty.longProperty("files.per.df.command", 100L);
            }
            numFiles = filesPerDfCommand.lookup(((OperatingSystemImpl)this.operatingSystem).getPlatformConfigBundle());
        }
        return this.map(fileSet, this.operatingSystem, numFiles.intValue());
    }
    
    protected Map<File, File> map(Set<File> fileSet, OperatingSystemImpl unixOperatingSystem, int filesPerCommand) {
        try {
            /*
             * Optimization: Only df the unique directories, as opposed to every file
             */
            Map<File, File> fileToDirectoryMap = mapFilesToCanonicalDirectories(fileSet);
            
            Map<File, File> directoryToMountpointMap = mapFilesToMountpoints(
                new LinkedHashSet<File>(fileToDirectoryMap.values()), unixOperatingSystem, filesPerCommand);
           
            return joinMaps(fileToDirectoryMap, directoryToMountpointMap);        
        }
        catch (IOException e) {
            /*
             * An IOException means that we could not determine the canonicalPath of a file.
             * Therefore, don't perform the directory optimization, and just df all of the files.
             */
            return mapFilesToMountpoints(fileSet, unixOperatingSystem, filesPerCommand);
        }
    }

    /**
     * Maps the given files to their canonical directories.
     * 
     * One entry in the returned Map will exist per file in the fileSet.
     * 
     * If a file is a directory itself, then the entry value will be the canonicalPath of the file.
     * 
     * If a file is not a directory, then the entry value will be the parent directory of the canonicalPath of the file.
     * @throws IOException if the canonical path of a file cannot be determined 
     * 
     */
    private Map<File, File> mapFilesToCanonicalDirectories(Set<File> fileSet) throws IOException {
        
        Map<File, File> fileToCanonicalDirectoryMap = new LinkedHashMap<File, File>();
        
        for (File file : fileSet) {
            File canonicalFile = file.getCanonicalFile();
            
            File canonicalDirectory;
            if (canonicalFile.isDirectory()) {
                canonicalDirectory = canonicalFile;
            }
            else {
                canonicalDirectory = canonicalFile.getParentFile();
            }
            fileToCanonicalDirectoryMap.put(file, canonicalDirectory);
            
        }
        
        return fileToCanonicalDirectoryMap;
    }

    /**
     * Maps the given files to their mountpoints.
     * 
     * One entry in the returned Map will exist per file in the fileSet.
     */
    private Map<File, File> mapFilesToMountpoints(
        Set<File> fileSet, 
        OperatingSystemImpl unixOperatingSystem,
        int filesPerCommand) {
        
        Map<File, File> fileToMountpointMap = new LinkedHashMap<File, File>();
        
        List<? extends List<File>> fileListBuckets = CollectionUtils.partition(new ArrayList<File>(fileSet), filesPerCommand);

        for (List<File> fileList : fileListBuckets) {
            DFResult dfResult = unixOperatingSystem.runDFCommand(fileList); 
        
            for (File file : fileList) {
                File mountpoint = dfResult.getFileInfo(file).getMountPoint();
                fileToMountpointMap.put(file, mountpoint);
            }            
        }
        
        return fileToMountpointMap;
    }

    /**
     * Joins the fileToDirectoryMap and the directoryToMountpointMap
     * into a single fileToMountpointMap.
     * 
     * One entry in the returned Map will exist per file in the fileToDirectoryMap.
     */
    private Map<File, File> joinMaps(Map<File, File> fileToDirectoryMap, Map<File, File> directoryToMountpointMap) {
        Map<File, File> fileToMountpointMap = new LinkedHashMap<File, File>();
        
        for (Map.Entry<File, File> fileToDirectoryMapEntry : fileToDirectoryMap.entrySet()) {
            fileToMountpointMap.put(fileToDirectoryMapEntry.getKey(), directoryToMountpointMap.get(fileToDirectoryMapEntry.getValue()));
        }
        
        return fileToMountpointMap;
    }

}
