/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/remote/RemoteObjectFactoryFactory.java#1 $
 * 
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.remote;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.Set;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.remote.jvm.JvmLaunchCommand;
import com.netapp.common.remote.jvm.JvmLaunchResult;
import com.netapp.common.util.GUID;
import com.netapp.common.util.Utils;

/**
 * Factory for obtaining {@link RemoteObjectFactory} instances.
 * 
 * Properties are exposed as setters for easy configuration via Spring:
 * <pre>
 *  &lt;bean id="remoteObjectFactoryFactory" class="com.netapp.common.remote.RemoteObjectFactoryFactory" 
 *        init-method="init" destroy-method="destroy"&gt;
 *       &lt;property name="registryPort" value="12345"/&gt;
 *       &lt;property name="createRegistry" value="true"/&gt;
 *  &lt;/bean&gt;
 * </pre>
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class RemoteObjectFactoryFactory {

    private static final NALogger logger = NALogger.getLogger();

    private int registryPort;

    private int spawnTimeout = 60 * 1000;
    
    private int spawnInterval = 1 * 1000;

    private String registryHost;
    
    private long lastSpawnTime;
    
    private RemoteObjectFactoryRegistry remoteObjectFactoryRegistry;

    private Set<GUID> spawnedRemoteObjectFactoryGUIDs = new HashSet<GUID>();

    /**
     * Specifies whether or not to create or reuse the remoteObjectFactoryRegistry on the given port 
     */
    private boolean createRegistry = true;    

    /**
     * @param createRegistry Specifies whether or not to create or reuse the remoteObjectFactoryRegistry on the given port 
     */
    public RemoteObjectFactoryFactory(int registryPort, boolean createRegistry) {
        this.registryPort = registryPort;
        this.createRegistry = createRegistry;
        this.init();
    }
    
    protected RemoteObjectFactoryFactory() {
    }
    
    public void init() {
        try {
            this.registryHost = InetAddress.getLocalHost().getHostAddress();
            
            if (this.createRegistry) {
                this.remoteObjectFactoryRegistry = new RemoteObjectFactoryRegistryImpl(this.registryPort);
                logger.internal("Started RMI Registry //%s:%d", this.registryHost, this.registryPort);
            }
            else {
                this.remoteObjectFactoryRegistry = RemoteObjectFactoryRegistryImpl.getRegistry(this.registryPort);
                logger.internal("Using RMI Registry //%s:%d", this.registryHost, this.registryPort);
            }
        }
        catch (UnknownHostException e) {
            throw new RuntimeException(e);
        }
        catch (RemoteException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Create a new RemoteObjectFactory instance in it's own JVM.
     * 
     * @return A reference to the newly created RemoteObjectFactory instance 
     */
    public RemoteObjectFactory createRemoteObjectFactory(RemoteObjectFactoryRequest remoteObjectFactoryRequest) {
        GUID guid = new GUID();
        JvmLaunchResult launchResult = this.spawnRemoteObjectFactoryJVM(remoteObjectFactoryRequest, guid);
        
        try {
            RemoteObjectFactory remoteObjectFactory = this.lookupRemoteObjectFactory(launchResult, guid);
            this.spawnedRemoteObjectFactoryGUIDs.add(guid);
            return remoteObjectFactory;
        }
        // CHECKSTYLE:OFF
        catch (RuntimeException e) {
            // CHECKSTYLE:ON
            /*
             * Make sure we don't leak the process if something went wrong
             */
            launchResult.destroy();
            throw e;
        }
    }
    
    private JvmLaunchResult spawnRemoteObjectFactoryJVM(RemoteObjectFactoryRequest remoteObjectFactoryRequest, GUID guid) {
        
        JvmLaunchCommand command = createJvmLaunchCommand(remoteObjectFactoryRequest, guid);

        /*
         * See burt 264642
         * Make sure the minimum amount of time between JVM spawns is greater
         * than the specified spawnInterval
         */
        synchronized (this) {
            while (true) {
                long currentTime = System.currentTimeMillis();
                if (currentTime - this.lastSpawnTime > this.spawnInterval) {
                    JvmLaunchResult jvmLaunchResult = remoteObjectFactoryRequest.getJvmLauncher().launch(command);
                    this.lastSpawnTime = System.currentTimeMillis();
                    return jvmLaunchResult;
                }
                else {
                    long delay = this.spawnInterval - (currentTime - this.lastSpawnTime);
                    logger.internal("Spawn of new RemoteObjectFactory requested within spawn interval - delaying %dms", delay);
                    Utils.sleep(delay);
                }
            }
        }
    }

    protected JvmLaunchCommand createJvmLaunchCommand(RemoteObjectFactoryRequest remoteObjectFactoryRequest, GUID guid) {
        return remoteObjectFactoryRequest.createJvmLaunchCommand(this.registryHost, 
            Integer.toString(this.registryPort), guid.toString());
    }
    
    private RemoteObjectFactory lookupRemoteObjectFactory(JvmLaunchResult jvmLaunchResult, GUID guid) {
        int msecsWaited = 0;
        while (msecsWaited < this.spawnTimeout) {
            try {
                RemoteObjectFactory remoteObjectFactory = (RemoteObjectFactory) this.remoteObjectFactoryRegistry.lookup(guid.toString());
                logger.internal("Successfully bound to RemoteObjectFactory with GUID %s", guid);
                return remoteObjectFactory;
            }
            catch (RemoteException e) {
                throw new RuntimeException(logger.error(MsgKey.REMOTE_UNABLE_TO_CONNECT_TO_SPAWNED_JVM_EXCEPTION, e, 
                                                          e.getLocalizedMessage()), e);
            }
            // Not yet registered - sleep and continue looping, until bound, timeout or spawned process dies on it's own
            catch (NotBoundException e) {
                logger.internal("Waiting for RemoteObjectFactory to be registered with GUID %s ...", guid);

                Utils.sleep(1000);
                msecsWaited += 1000;

                // Make sure the process that is supposed to registering the key has not died on it's own
                // The timeout can be long and this lets us fail fast.
                if (!jvmLaunchResult.isRunning()) {
                    throw new RuntimeException(logger.error(MsgKey.REMOTE_UNABLE_TO_CONNECT_TO_SPAWNED_JVM_EXCEPTION,
                        jvmLaunchResult));
                }
            }
        }

        // Timed out waiting on the JVM to start properly
        throw new RuntimeException(logger.error(MsgKey.REMOTE_UNABLE_TO_CONNECT_TO_SPAWNED_JVM_TIMEOUT, this.spawnTimeout / 1000.0, jvmLaunchResult));
    }
    
    public void setCreateRegistry(boolean createRegistry) {
        this.createRegistry = createRegistry;
    }
    
    /**
     * @param registryPort port to be used for this JVM's internal RMI remoteObjectFactoryRegistry
     */
    public void setRegistryPort(int registryPort) {
        this.registryPort = registryPort;
    }
 
    public void setSpawnTimeout(int spawnTimeout) {
        this.spawnTimeout = spawnTimeout;
    }
    
    public void setSpawnInterval(int spawnInterval) {
        this.spawnInterval = spawnInterval;
    }

    public void destroy() {
        this.destroyAllSpawnedRemoteObjectFactories();
        this.destroyRMIRegistry();
    }
    
    public void destroyAllSpawnedRemoteObjectFactories() {
        logger.internal("Destroying all spawned RemoteObjectFactories");
        for (GUID guid : this.spawnedRemoteObjectFactoryGUIDs) {
            try {
                RemoteObjectFactory remoteObjectFactory = (RemoteObjectFactory) this.remoteObjectFactoryRegistry.lookup(guid.toString());
                remoteObjectFactory.destroy();
            }
            catch (RemoteException e) {
                logger.internalTrace("Ignoring exception while destroying RemoteObjectFactory.", e);
            }
            catch (NotBoundException e) {
                logger.internalTrace("Ignoring exception while destroying RemoteObjectFactory.", e);
            }
        }
    }
    
    void destroyRMIRegistry() {
        if (this.createRegistry && this.remoteObjectFactoryRegistry != null) {
            logger.internal("Destroying RMI Registry //%s:%d", this.registryHost, this.registryPort);
            try {
                this.remoteObjectFactoryRegistry.destroy();
                this.remoteObjectFactoryRegistry = null;
            }
            catch (RemoteException e) {
                logger.internal("Ignoring exception while destroying RemoteObjectFactoryRegistry.", e);
            }
        }
    }
    
    @Override
    protected void finalize() throws Throwable {
        this.destroyRMIRegistry();
        super.finalize();
    }
}
