/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/spring/SpringInstanceFactory.java#1 $
 * 
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.spring;

import java.util.Collection;
import java.util.Map;

import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;


/**
 * Find instances of a given class in the Spring context scope, or
 * create a new instance if not defined.
 * 
 * In most cases, the defined beans should be prototype scope.
 * 
 * @see com.netapp.common.flow.SpringTaskInstanceFactory
 * 
 * @author sirianni, sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class SpringInstanceFactory<T> implements ApplicationContextAware {
    private ApplicationContext applicationContext;

    public void setApplicationContext(ApplicationContext applicationContext) {
        this.applicationContext = applicationContext;
    }

    /**
     * Returns the single bean whose class equals the given class, if there is one.
     * If there is no instance defined in Spring, create a new instance reflectively
     * without setting any properties.
     * 
     * @param clazz Class of bean to find
     * @return new instance of clazz
     * @throws InstantiationException
     * @throws IllegalAccessException
     */
    public T newInstance(Class<? extends T> clazz) throws InstantiationException, IllegalAccessException {
        // Get beans that are instances of the given taskClass or its subclasses.
        Map<?, ?> beans = this.applicationContext.getBeansOfType(clazz);
        
        T obj = null;
        if (beans.size() == 1) {
            obj = clazz.cast(beans.values().iterator().next());
        }
        else {
            // If we find more than one, try to pick the one whose class is the exact
            // same as taskClass (rather than an instance of a subclass).
            // This lets us have Spring managed tasks who are subclasses of other
            // Spring managed tasks.
            obj = filterByConcreteClass(beans.values(), clazz);
        }
        
        if (obj != null) {
            return returningSpringInstantiatedBean(obj);
        }
        return returningReflectedInstantiatedBean(clazz.newInstance());
    }


    /**
     * Returns the single bean whose class equals the given class, if there is one.
     * Returns null if no bean is found whose class equals the given task class OR 
     * multiple beans are found whose class equals the given task class.
     */
    private T filterByConcreteClass(Collection<?> beans, Class<? extends T> clazz) {
        Object concreteBean = null;
        for (Object bean : beans) {
            if (bean.getClass().equals(clazz)) {
                if (concreteBean == null) {
                    concreteBean = bean;
                }
                else {
                    // multiple beans found 
                    return null;
                }
            }
        }
        return clazz.cast(concreteBean);
    }
    
    /**
     * Called immediately prior to returning a bean discovered in the Spring container.
     * Subclasses may override this hook in order to perform post-processing
     * @param bean
     * @return bean
     */
    protected T returningSpringInstantiatedBean(T bean) {
        return bean;
    }
    
    /**
     * Called immediately prior to returning a bean reflectively created that wasn't in
     * the Spring container. Subclasses may override this hook in order to perform
     * post-processing.
     * @param bean
     * @return bean
     */
    protected T returningReflectedInstantiatedBean(T bean) {
        return bean;
    }
}
