/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/text/SubstitutionPattern.java#1 $
 *
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.text;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.util.CollectionUtils;

/**
 * Performs string substitution for strings with variables in the form {variablename}.
 * 
 * Supports:
 * 
 * Testing for invalid variables
 * Testing for required variables
 * Changing the start and end variable tokens (may be needed for different languages)  
 * 
 * @author philipc
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class SubstitutionPattern {
    
    public static interface SubstitutionPatternElement {
        String getValue(SubstitutionPatternContext context);
    }
    
    public static final class TextElement implements SubstitutionPatternElement {
        
        private String text;
        
        public TextElement(String text) {
            this.text = text;
        }

        public String getValue(SubstitutionPatternContext context) {
            return this.text;
        }
        
        @Override
        public String toString() {
            return this.text;
        }
        
        @Override
        public boolean equals(Object obj) {
            return EqualsBuilder.reflectionEquals(this, obj);
        }
        
        @Override
        public int hashCode() {
            return HashCodeBuilder.reflectionHashCode(this);
        }
        
    }

    public static final class VariableElement implements SubstitutionPatternElement {
        
        private String name;
        private final String variableStartToken;
        private final String variableEndToken;
        
        public VariableElement(String name) {
            this(name, getDefaultVariableStartToken(), getDefaultVariableEndToken());
        }
        
        public VariableElement(String name, String variableStartToken, String variableEndToken) {
            this.name = name;
            this.variableStartToken = variableStartToken;
            this.variableEndToken = variableEndToken;
        }
        
        public String getName() {
            return this.name;
        }

        public String getValue(SubstitutionPatternContext context) {
            return context.getValue(this.name);
        }
        
        @Override
        public String toString() {
            return this.variableStartToken + this.name + this.variableEndToken;
        }
        
        @Override
        public boolean equals(Object obj) {
            return EqualsBuilder.reflectionEquals(this, obj);
        }
        
        @Override
        public int hashCode() {
            return HashCodeBuilder.reflectionHashCode(this);
        }
        
    }
    
    private final List<SubstitutionPatternElement> elements;
    
    /**
     * Use {@link #compile(String, Set, Set)} to construct.
     * 
     * @param elements
     * @param variableEndToken 
     * @param variableStartToken 
     */
    private SubstitutionPattern(List<SubstitutionPatternElement> elements) {
        this.elements = Collections.unmodifiableList(elements);
    }
    
    /**
     * Returns the string with all variables substituted with 
     * the values provided by the given context.
     * 
     */
    public String render(SubstitutionPatternContext context) {
        StringBuilder builder = new StringBuilder();
        for (SubstitutionPatternElement element : this.elements) {
            builder.append(element.getValue(context));
        }
        return builder.toString();
    }
    
    /**
     * Returns the raw string pattern with no variables substituted.
     *
     */
    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder();
        for (SubstitutionPatternElement element : this.elements) {
            builder.append(element.toString());
        }
        return builder.toString();
    }
    
    public List<SubstitutionPatternElement> getElements() {
        return this.elements;
    }
    
    public Set<String> getVariableNames() {
        Set<String> variableNames = new HashSet<String>();
        for (SubstitutionPatternElement element : this.elements) {
            if (element instanceof VariableElement) {
                variableNames.add(((VariableElement) element).getName());
            }
        }
        return variableNames;
    }
    
    /**
     * 
     * Compiles the given patternString into a SubstitutionPattern using the default
     * variable start and end tokens as defined by the resource bundle.
     * 
     * @param patternString pattern string containing {variables}
     * @param validVariableNames names of valid variables.
     *                           if other variables are encountered a SubstitutionPatternSyntaxException is thrown.
     *                           if null, then no is valid checking is performed
     * @param requiredVaribleNames names of required variables
     *                           if these variables are not encountered a SubstitutionPatternSyntaxException is thrown.
     *                           if null, then no is required checking is performed
     * @return Substitution pattern for the supplied elements
     * @throws SubstitutionPatternSyntaxException
     */
    public static SubstitutionPattern compile(
        String patternString, 
        Set<String> validVariableNames, 
        Set<String> requiredVaribleNames) {
        
        return compile(
            patternString, 
            validVariableNames, 
            requiredVaribleNames,
            getDefaultVariableStartToken(),
            getDefaultVariableEndToken());
    }

    /**
     * 
     * Compiles the given patternString into a SubstitutionPattern.
     * 
     * @param patternString pattern string containing {variables}
     * @param validVariableNames names of valid variables.
     *                           if other variables are encountered a SubstitutionPatternSyntaxException is thrown.
     *                           if null, then no is valid checking is performed
     * @param requiredVaribleNames names of required variables
     *                           if these variables are not encountered a SubstitutionPatternSyntaxException is thrown.
     *                           if null, then no is required checking is performed
     * @param variableStartToken string that denotes the start of a variable (typically '{')
     * @param variableEndToken  string that denotes the end of a variable (typically '}')
     * @return SubstitutionPattern for the supplied elements
     * @throws SubstitutionPatternSyntaxException
     */
    public static SubstitutionPattern compile(
        String patternString, 
        Set<String> validVariableNames, 
        Set<String> requiredVaribleNames,
        String variableStartToken,
        String variableEndToken) {
        
        List<SubstitutionPatternElement> elements = new ArrayList<SubstitutionPatternElement>();
        
        int start = 0;
        int variableStartIndex = 0;
        while (start < patternString.length() && (variableStartIndex = patternString.indexOf(variableStartToken, start)) != -1) {
            
            int variableEndIndex = patternString.indexOf(variableEndToken, variableStartIndex);
            
            if (variableEndIndex != -1) {
                
                if (variableStartIndex != start) {
                    elements.add(new TextElement(patternString.substring(start, variableStartIndex)));
                }
                
                String variableName = patternString.substring(variableStartIndex + variableStartToken.length(), variableEndIndex);
                elements.add(new VariableElement(variableName, variableStartToken, variableEndToken));
                start = variableEndIndex + 1;
            }
            else {
                elements.add(new TextElement(patternString.substring(start)));
                start = patternString.length();
            }
            
        }

        if (start < patternString.length()) {
            elements.add(new TextElement(patternString.substring(start)));
        }
        
        return compile(elements, validVariableNames, requiredVaribleNames);
        
    }

    /**
     * 
     * Compiles the given elements into a SubstitutionPattern.
     * 
     * @param elements list of SubstitutionPatternElements
     * @param validVariableNames names of valid variables.
     *                           if other variables are encountered a SubstitutionPatternSyntaxException is thrown.
     *                           if null, then no is valid checking is performed
     * @param requiredVaribleNames names of required variables
     *                           if these variables are not encountered a SubstitutionPatternSyntaxException is thrown.
     *                           if null, then no is required checking is performed
     * @return SubstitutionPattern for the supplied elements
     * @throws SubstitutionPatternSyntaxException
     */
    public static SubstitutionPattern compile(
        List<SubstitutionPatternElement> elements, 
        Set<String> validVariableNames, 
        Set<String> requiredVaribleNames) {
        
        Set<String> requiredVaribleNamesRemaining = requiredVaribleNames == null ? new HashSet<String>() : new HashSet<String>(requiredVaribleNames);
        Set<String> invalidVaribleNames = new HashSet<String>();
        
        StringBuilder patternString = new StringBuilder();

        for (SubstitutionPatternElement element : elements) {
            
            if (element instanceof VariableElement) {
                VariableElement variableElement = (VariableElement) element;
                
                if (validVariableNames != null && !validVariableNames.contains(variableElement.getName())) {
                    invalidVaribleNames.add(variableElement.getName());
                }
                
                requiredVaribleNamesRemaining.remove(variableElement.getName());
            }
            
            patternString.append(element.toString());
        }

        if (!requiredVaribleNamesRemaining.isEmpty()) {
            throw new SubstitutionPatternSyntaxException(
                MessageBundle.localizeMessage(MsgKey.SUBSTITUTION_PATTERN_REQUIRED_VARIABLES_NOT_FOUND, 
                    CollectionUtils.collectionToString(requiredVaribleNamesRemaining, ", ", "{", "}"), 
                    CollectionUtils.collectionToString(requiredVaribleNamesRemaining, "_", "{", "}"), 
                    patternString), true, false);
        }
        
        if (!invalidVaribleNames.isEmpty()) {
            throw new SubstitutionPatternSyntaxException(
                MessageBundle.localizeMessage(MsgKey.SUBSTITUTION_PATTERN_INVALID_VARIABLES, 
                    CollectionUtils.collectionToString(invalidVaribleNames, ", "), 
                    patternString), false, true);
        }
        
        return new SubstitutionPattern(elements);
        
    }

    public static String getDefaultVariableStartToken() {
        return MessageBundle.localizeMessage(MsgKey.SUBSTITUTION_PATTERN_VARIABLE_START_TOKEN);
    }

    public static String getDefaultVariableEndToken() {
        return MessageBundle.localizeMessage(MsgKey.SUBSTITUTION_PATTERN_VARIABLE_END_TOKEN);
    }

}
