/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/Interval.java#1 $
 *
 * Copyright (c) 2009 NetApp, Inc.
 * All rights reserved. 
 */
package com.netapp.common.util;

import static java.lang.Math.abs;

import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.google.common.collect.ImmutableMap;

/**
 * Utility class to represent time intervals and convert back and forth to
 * string values.
 * 
 * This is largely borrowed from and intended to be compatible with the
 * interval_* functions in libadt/utils.c.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class Interval implements Comparable<Interval> {
    /**
     * Pattern to match time strings of "[dd:]hh:mm:ss"
     */
    private static final Pattern TIME_STRING_PATTERN =
        Pattern.compile("(-)?\\s*(?:(\\d+):)?(\\d+):(\\d+):(\\d+)");

    private static final int TIME_STRING_SIGN_GROUP = 1;
    private static final int TIME_STRING_DAY_GROUP = 2;
    private static final int TIME_STRING_HOUR_GROUP = 3;
    private static final int TIME_STRING_MIN_GROUP = 4;
    private static final int TIME_STRING_SEC_GROUP = 5;
    
    /**
     * Pattern to match time strings of "XX unit"
     */
    private static final Pattern TIME_SUFFIX_PATTERN =
        Pattern.compile("(-?\\d+(?:\\.\\d*)?)\\s*(\\w+)");
    private static final int TIME_SUFFIX_VALUE_GROUP = 1;
    private static final int TIME_SUFFIX_UNIT_GROUP = 2;

    public static final String[] SECOND_UNITS = {"s", "sec", "secs", "seond", "seconds"};
    public static final double SECOND_MUTLIPLIER = 1.0;
    
    public static final String[] MINUTE_UNITS = {"m", "min", "mins", "minute", "minutes"};
    public static final double MINUTE_MULTIPLIER = 60.0 * SECOND_MUTLIPLIER;
    
    public static final String[] HOUR_UNITS = {"h", "hr", "hrs", "hour", "hours"};
    public static final double HOUR_MULTIPLIER = 60.0 * MINUTE_MULTIPLIER;
    
    public static final String[] DAY_UNITS = {"d", "day", "days"};
    public static final double DAY_MULTIPLIER = 24.0 * HOUR_MULTIPLIER;
    
    public static final String[] WEEK_UNITS = {"w", "wk", "week", "weeks"};
    public static final double WEEK_MULTIPLIER = 7.0 * DAY_MULTIPLIER;

    private static final Map<String, Double> TIME_MULTIPLIERS = new ImmutableMap.Builder<String, Double>()
        .put("s", SECOND_MUTLIPLIER)
        .put("sec", SECOND_MUTLIPLIER)
        .put("second", SECOND_MUTLIPLIER)
        .put("seconds", SECOND_MUTLIPLIER)
        .put("m", MINUTE_MULTIPLIER)
        .put("min", MINUTE_MULTIPLIER)
        .put("mins", MINUTE_MULTIPLIER)
        .put("minute", MINUTE_MULTIPLIER)
        .put("minutes", MINUTE_MULTIPLIER)
        .put("h", HOUR_MULTIPLIER)
        .put("hr", HOUR_MULTIPLIER)
        .put("hrs", HOUR_MULTIPLIER)
        .put("hour", HOUR_MULTIPLIER)
        .put("hours", HOUR_MULTIPLIER)
        .put("d", DAY_MULTIPLIER)
        .put("day", DAY_MULTIPLIER)
        .put("days", DAY_MULTIPLIER)
        .put("w", WEEK_MULTIPLIER)
        .put("wk", WEEK_MULTIPLIER)
        .put("wks", WEEK_MULTIPLIER)
        .put("week", WEEK_MULTIPLIER)
        .put("weeks", WEEK_MULTIPLIER)
        .build();

    private final long seconds;
    
    /**
     * Construct an interval of a number of seconds
     * @param seconds
     */
    public Interval(long seconds) {
        this.seconds = seconds;
    }
    
    /**
     * Construct an interval from a time string.
     * @param value
     * @throws IllegalArgumentException The supplied value could not be converted to an interval
     * @see #parse
     */
    public Interval(String value) {
        this.seconds = parse(value);
    }
    
    /**
     * Get the number of seconds represented by this interval
     * @return interval value
     */
    public long getSeconds() {
        return this.seconds;
    }
    
    /**
     * Get the number of milliseconds represented by this interval
     * @return interval value in ms
     */
    public long getMillis() {
        return this.seconds * 1000;
    }
    
    /**
     * Convert this interval value to a string.
     * 
     * String is created in suffix notation using 'seconds',
     * 'minutes', 'hours', or 'days' to specify the units.
     * @return Suffix style representation of interval
     */
    public String getSuffixString() {
        double dval;
        String units;
        String str;

        /*
         *  Select appropriate units so that dval is as
         *  small as possible but dval >= 1.0
         */
        if (abs(seconds) < 60) {
            units = "second";
            dval = (double) seconds;
        } 
        else if (abs(seconds) < 3600) {
            units = "minute";
            dval = (double) seconds / MINUTE_MULTIPLIER;
        } 
        else if (abs(seconds) < 86400) {
            units = "hour";
            dval = (double) seconds / HOUR_MULTIPLIER;
        } 
        else if (abs(seconds) < 604800) {
            units = "day";
            dval = (double) seconds / DAY_MULTIPLIER;
        } 
        else {
            units = "week";
            dval = (double) seconds / WEEK_MULTIPLIER;
        }

        /*
         *  Only display as many decimal places as needed
         *  and never display more than 2.
         *  If not exactly 1 unit, make noun plural.
         *  
         *  This is ripped directly from libadt. Author is not
         *  responsible for floating point bugs herein.
         */
        if (1.0 == dval) {
            str = String.format("1 %s", units);
        } 
        else if (0 == dval - (double) ((int) dval)) {
            str = String.format("%.0f %ss", dval, units);
        } 
        else if (0 == dval * 10.0 - (double) ((int) (dval * 10.0))) {
            str = String.format("%.1f %ss", dval, units);
        } 
        else {
            str = String.format("%.2f %ss", dval, units);
        }

        return str;
    }
    
    /**
     * Convert this interval to a string in timestamp [DD:]HH:MM:SS format.
     * @return timestamp format string
     */
    public String getTimeString() {
        String str;
        if (seconds <= -86400) {
            str = String.format("-%d:%02d:%02d:%02d",
                -seconds / 86400,
                (-seconds / 3600) % 24,
                (-seconds / 60) % 60,
                -seconds % 60);
        } 
        else if (seconds < 0) {
            str = String.format("-%02d:%02d:%02d",
                -seconds / 3600,
                (-seconds / 60) % 60,
                -seconds % 60);
        } 
        else if (seconds < 86400) {
            str = String.format("%02d:%02d:%02d",
                seconds / 3600,
                (seconds / 60) % 60,
                seconds % 60);
        } 
        else {
            str = String.format("%d:%02d:%02d:%02d",
                seconds / 86400,
                (seconds / 3600) % 24,
                (seconds / 60) % 60,
                seconds % 60);
        }

        return str;
    }
    
    @Override
    public String toString() {
        return getTimeString();
    }
    
    /**
     * Parse a string interval representation into seconds.
     * Handles both suffix style (xx units) and interval
     * ([DD:]HH:MM:SS) style inputs.
     * @param value
     * @return Duration of interval in seconds
     * @throws IllegalArgumentException The supplied value could not be converted to an interval
     */
    public static long parse(String value) {
        Matcher matcher = TIME_STRING_PATTERN.matcher(value);
        if (matcher.matches()) {
            int sign = 1;
            int d = 0;
            int h = 0;
            int m = 0;
            int s = 0;
            if (matcher.group(TIME_STRING_SIGN_GROUP) != null) {
                sign = -1;
            }
            if (matcher.group(TIME_STRING_DAY_GROUP) != null) {
                d = Integer.parseInt(matcher.group(TIME_STRING_DAY_GROUP), 10);
            }
            h = Integer.parseInt(matcher.group(TIME_STRING_HOUR_GROUP));
            m = Integer.parseInt(matcher.group(TIME_STRING_MIN_GROUP));
            s = Integer.parseInt(matcher.group(TIME_STRING_SEC_GROUP));
            return sign * (s + 60L * (m + 60L * (h + 24L * d)));       
        }
        matcher = TIME_SUFFIX_PATTERN.matcher(value);
        if (matcher.matches()) {
            double dval = Double.parseDouble(matcher.group(TIME_SUFFIX_VALUE_GROUP));
            int ival = 0;
            String suffix = matcher.group(TIME_SUFFIX_UNIT_GROUP).toLowerCase();
            
            Double mult = TIME_MULTIPLIERS.get(suffix);
            if (mult != null) {
                dval = dval * mult;
                if (dval < 0.0) {
                    ival = (int)(dval - 0.5);
                }
                else {
                    ival = (int)(dval + 0.5);
                }
                return ival;
            }
            else {
                throw new IllegalArgumentException(value);
            }
        }
        throw new IllegalArgumentException(value);
    }
    
    @Override
    public int hashCode() {
        return (int)(seconds ^ (seconds >>> 32));
    }
    
    @Override
    public boolean equals(Object o) {
        return (o instanceof Interval)
            && (((Interval)o).seconds == this.seconds);
    }
    
    public int compareTo(Interval o) {
        long thisVal = this.seconds;
        long anotherVal = o.seconds;
        return (thisVal < anotherVal ? -1 : (thisVal == anotherVal ? 0 : 1));
    }
}
