/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/Version.java#1 $
 * 
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * Encapsulates the concept of a version.
 * @author rippy
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class Version implements Comparable<Version> {
    
    private String versionString;

    public Version(final String versionString) {
        this.versionString = versionString;
    }

    /**
     * Compare this Version to another.
     *
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(Version other) {
        return compareTo(other, 0);
    }
    
    /**
     * Compare this Version to another to the specified signficance.
     * i.e. 1.0.0 == 1.0.2 to two places while 1.0.0 != 1.1.0.
     * 
     * @param other Other Version to compare
     * @param significance Significance to compare to. 0 compares the entire string.
     */
    public int compareTo(Version other, int significance) {
        if (this.versionString.equalsIgnoreCase(other.versionString)) {
            return 0;
        }

        List<String> myVersionParts = this.split();
        List<String> otherVersionParts = other.split();
        
        /*
         * Pad them both to the same length with 0s to handle the "1.0.0" == "1.0" case
         */
        int longerLength = Math.max(myVersionParts.size(), otherVersionParts.size());
        if (myVersionParts.size() != otherVersionParts.size()) {

            List<String> shorterList = (myVersionParts.size() < otherVersionParts.size()) ? myVersionParts : otherVersionParts;
            
            do {
                shorterList.add("0");
            } 
            while (shorterList.size() < longerLength);
        }
        
        /*
         * Check if the user has constrained how far they're interested in looking
         */
        if ((significance > 0) && significance < longerLength) {
            longerLength = significance;
        }
        
        for (int i = 0; i < longerLength; i++) {
            int compare = myVersionParts.get(i).compareTo(otherVersionParts.get(i));
            if (compare != 0) {
                return compare;
            }
        }

        // no differences detected
        return 0;
    }

    private List<String> split() {
        List<String> myVersionParts = new ArrayList<String>();
        StringTokenizer st = new StringTokenizer(this.versionString, ".");
        while (st.hasMoreTokens()) {
            myVersionParts.add(st.nextToken());
        }
        return myVersionParts;
    }
    
    /**
     * Is this version equivalent to the other version?
     * 
     * Note: Slightly different semantics than {@link #equals(Object)}.  This
     * method will return true for "3.0.0" and "3.0" wheras {@link #equals(Object)} 
     * will not.
     */
    public boolean isEquivalentTo(Version other) {
        return compareTo(other) == 0;
    }

    public boolean isGreaterThan(Version other) {
        return compareTo(other) > 0;
    }
    
    public boolean isGreaterThanOrEqualTo(Version other) {
        return compareTo(other) >= 0;
    }

    public boolean isLessThan(Version other) {
        return compareTo(other) < 0;
    }
    
    public boolean isLessThanOrEqualTo(Version other) {
        return compareTo(other) <= 0;
    }
    
    @Override
    public String toString() {
        return this.versionString;
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof Version)) {
            return false;
        }
        
        Version other = (Version) obj;
        return this.versionString.equals(other.versionString);
    }
    
    @Override
    public int hashCode() {
        return this.versionString.hashCode();
    }
    
}
