/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/fault/FaultInjectionManager.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util.fault;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.StringUtils;

/**
 * Provide a hook to inject replacement classes in an application. This is generally
 * designed for fault injection testing, hence the name. At an application site where
 * classes may be instantiated via <code>Class.newInstance()</code>, use {@link #createObject(Class, String...)}
 *  using the original class object instead to create a new object.<p>
 * To setup at runtime, create replacement classes and a fault map in a separate classpath, 
 * and supply this classpath to the JVM in the system property {@link #SYSTEM_PROPERTY}.<p>
 * Faults are defined in the file 'fault.properties', which takes the following format:
 * &lt;key&gt;.&lt;original_class&gt;=&lt;method&gt:&lt;class&gt;
 * <ul>
 *   <li><code>key</code> is an application specific code defined for the replacement site
 *   <li><code>original_class</code> is the original class to replace
 *   <li><code>method</code> is used to interpret the new class name as described below
 *   <li><code>class</code> is the name of the new class to load from the injection classpath
 * </ul>
 * Method specifies how to derive the new class:
 * <ul>
 *   <li><code>class</code> will load the specified class and return it directly. The replacement class
 *   must be a subclass of the original class.
 *   <li><code>factory</code> will load the specified class which must be an instance of {@link FaultFactory}.
 *   An instance of the class will be constructed and the replacement class retrieved via the 
 *   {@link FaultFactory#createObject(Class, ClassLoader)} method.
 * </ul>
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class FaultInjectionManager {
    private static final NALogger logger = NALogger.getLogger(FaultInjectionManager.class);
    private static final Pattern SPEC_REGEX = Pattern.compile("([^:]+):(.+)");
    
    /** The direct class replacement method */
    public static final String CLASS_METHOD = "class";
    /** The factory replacement method */
    public static final String FACTORY_METHOD = "factory";
    
    /** The system property used to enable fault injection */
    public static final String SYSTEM_PROPERTY = "com.netapp.common.util.FaultClassPath";
    
    /** The fault property map file */
    public static final String FAULT_MAP_FILE = "faults.properties";
    
    private static final FaultInjectionManager INSTANCE;
    static {
        FaultInjectionManager manager = null;
        String classpath = System.getProperty(SYSTEM_PROPERTY);
        if (!StringUtils.isEmpty(classpath)) {
            String[] paths = classpath.split(File.pathSeparator);
            try {
                manager = new FaultInjectionManager(paths, FAULT_MAP_FILE);
                logger.info(MsgKey.UTIL_FAULT_INITIALIZED);
            }
            // CHECKSTYLE:OFF
            catch (Exception e) {
            // CHECKSTYLE:ON
                logger.error(MsgKey.UTIL_FAULT_ERROR_INITIALIZING, e, e.getLocalizedMessage());
            }
        }
        INSTANCE = manager;
    }
    
    private final Properties configuration;
    private final URLClassLoader baseLoader;
    private final Map<ClassLoader, ClassLoader> chainedLoaders;
    
    /**
     * Construct a new FaultInjectionManager
     * @param paths Paths to use for the injected classpath
     * @param propertiesFile Replacement map file
     * @throws IOException
     */
    public FaultInjectionManager(String[] paths, String propertiesFile) throws IOException {
        this.configuration = new Properties();
        this.chainedLoaders = new HashMap<ClassLoader, ClassLoader>();
        
        URL[] urls = new URL[paths.length];
        
        for (int i = 0; i < paths.length; i++) {
            urls[i] = new File(paths[i]).toURI().toURL();
        }
        this.baseLoader = new URLClassLoader(urls, null);
        
        InputStream stream = this.baseLoader.getResourceAsStream(propertiesFile);
        if (stream != null) {
            this.configuration.load(stream);
        }
        else {
            throw new FileNotFoundException(propertiesFile);
        }
    }
    
    /**
     * Lookup the supplied class and site key(s), and retrieve the
     * configured replacement class.
     * @param <T>
     * @param clazz Original class
     * @param keys One or more application defined keys defining the site and/or context to replace with
     * @return Subclass defined by this manager or the original class
     * @throws IllegalAccessException 
     * @throws InstantiationException 
     */
    public <T> T createObject(Class<T> clazz, String... keys) throws InstantiationException, IllegalAccessException {
        for (String key : keys) {
            if (key == null) {
                continue;
            }
            String fullKey = key + "." + clazz.getName();
            logger.debug(MsgKey.UTIL_FAULT_KEY_LOOKUP, fullKey);
            String value = this.configuration.getProperty(fullKey);
            if (StringUtils.isEmpty(value)) {
                continue;
            }
            
            T result = createReplacementObject(clazz, value);
            if (result != null) {
                return result;
            }
        }
        return clazz.newInstance();
    }
    
    /**
     * Get a class loader configured with our classpath. Multiple calls to this method with
     * the same parent will return the same class loader.
     * @param parent Parent classloader
     * @return Classloader for our configured path with parent defined
     */
    public ClassLoader getClassLoader(ClassLoader parent) {
        ClassLoader loader = this.chainedLoaders.get(parent);
        if (loader == null) {
            loader = new URLClassLoader(this.baseLoader.getURLs(), parent);
            this.chainedLoaders.put(parent, loader);
        }
        return loader;
    }
    
    /**
     * Load a replacement class according to the supplied spec
     * @param <T>
     * @param clazz Original class
     * @param spec Spec of the format "method:NewClass"
     * @return Newly loaded class, or null if the replacement class could not be loaded
     * @throws IllegalAccessException 
     * @throws InstantiationException 
     */
    private <T> T createReplacementObject(Class<T> clazz, String spec) throws InstantiationException, IllegalAccessException {
        Matcher m = SPEC_REGEX.matcher(spec);
        if (!m.matches()) {
            logger.error(MsgKey.UTIL_FAULT_BAD_SPEC, spec, clazz.getName());
            return null;
        }
        String method = m.group(1);
        String newClassName = m.group(2);
        ClassLoader loader = getClassLoader(clazz.getClassLoader());
        Class<?> newClass = null;
        try {
            newClass = loader.loadClass(newClassName);
        }
        catch (ClassNotFoundException e) {
            logger.error(MsgKey.UTIL_FAULT_CLASS_NOT_FOUND, clazz.getName(), newClassName);
            return null;
        }
        
        if (CLASS_METHOD.equalsIgnoreCase(method)) {
            if (!clazz.isAssignableFrom(newClass)) {
                logger.error(MsgKey.UTIL_FAULT_CLASS_NOT_SUBCLASS, clazz.getName(), newClass.getName());
                return null;
            }

            logger.info(MsgKey.UTIL_FAULT_INJECTING_CLASS, clazz.getName(), newClass.getName());
            return newClass.asSubclass(clazz).newInstance();
        }
        else if (FACTORY_METHOD.equalsIgnoreCase(method)) {
            if (!FaultFactory.class.isAssignableFrom(newClass)) {
                logger.error(MsgKey.UTIL_FAULT_FACTORY_NOT_INSTANCE, clazz.getName(), newClass.getName());
                return null;
            }
            FaultFactory factory;
            try {
                logger.info(MsgKey.UTIL_FAULT_INJECTING_FACTORY_CLASS, clazz.getName(), newClass.getName());
                factory = newClass.asSubclass(FaultFactory.class).newInstance();
            }
            catch (InstantiationException e) {
                logger.error(MsgKey.UTIL_FAULT_FACTORY_INSTANTIATION_ERROR, e, newClass.getName(), e.getLocalizedMessage());
                return null;
            }
            catch (IllegalAccessException e) {
                logger.error(MsgKey.UTIL_FAULT_FACTORY_INSTANTIATION_ERROR, e, newClass.getName(), e.getLocalizedMessage());
                return null;
            }
            return factory.createObject(clazz, loader);
        }
        
        logger.error(MsgKey.UTIL_FAULT_BAD_SPEC_METHOD, method, clazz.getName());
        return null;
    }
    
    /**
     * Retrieve the system configured instance
     * @return FaultInjectionManager set by {@link #SYSTEM_PROPERTY} or null if
     * not configured.
     */
    public static FaultInjectionManager getInstance() {
        return INSTANCE;
    }
    
    /**
     * Check if a manager was configured by {@link #SYSTEM_PROPERTY}
     * @return True if a FaultInjectionManager instance exists
     */
    public static boolean isActive() {
        return INSTANCE != null;
    }
}
