/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/generator/autozapi/AutoZapiTypeGenerator.java#1 $
 * 
 * Copyright (c) 2009 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.generator.autozapi;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import com.netapp.common.generator.NetappClassGenerator;
import com.netapp.common.zapi.generator.GeneratorUtils;
import com.netapp.common.zapi.generator.TypeGenerator;
import com.netapp.common.zapi.parser.Field;
import com.netapp.common.zapi.parser.Type;

public abstract class AutoZapiTypeGenerator<T extends Type> extends TypeGenerator<T> {
    
    /** Package containing the ZAPIRunner annotations */
    public static final String ANNOTATION_PACKAGE = "com.netapp.common.zapi.annotation";
    
    private String zapiTypePrefix;
    
    public AutoZapiTypeGenerator(AutoZapiGeneratorFactory generatorFactory, T type) {
        super(generatorFactory, type);
        zapiTypePrefix = AutoZapiGeneratorFactory.PROPERTY_ZAPI_TYPE_PREFIX.lookup(generatorFactory.getProperties());
    }
    
    @Override
    public String getNameGenerated() {
        String name = GeneratorUtils.zapiToUpperCamelCase(getTarget().getName());
        return (zapiTypePrefix != null) ? (zapiTypePrefix + name) : name;
    }
    
    public String getUnprefixedNameGenerated() {
        return GeneratorUtils.zapiToUpperCamelCase(getTarget().getName());
    }
    
    /**
     * Get the java package this type is in.
     * @return
     */
    public String getPackageName() {
        AutoZapiNamespaceGenerator namespaceGenerator = (AutoZapiNamespaceGenerator) getGeneratorFactory().getNamespaceGenerator(getTarget().getNamespace());
        AutoZapiCategoryGenerator categoryGenerator = (AutoZapiCategoryGenerator) getGeneratorFactory().getCategoryGenerator(getTarget().getCategory());
        return namespaceGenerator.getPackageName() + "." + categoryGenerator.getNameGenerated();
    }
    
    /**
     * Get the complete java qualified name of this class
     * @return
     */
    public String getQualifiedName() {
        return getPackageName() + "." + getNameGenerated();
    }
    
    /**
     * Get the list of class imports required for the type
     * and any defined fields, including java types and annotations.
     * @return
     */
    @SuppressWarnings("unchecked")
    public Set<String> getImports(Collection<Field> fields) {
        Set<String> imports = new HashSet<String>();
        imports.add(ANNOTATION_PACKAGE + "." + getTarget().getClass().getSimpleName());
        for (Field f : fields) {
            imports.addAll(((AutoZapiFieldGenerator) getGeneratorFactory().getFieldGenerator(f)).getImports());
        }
        return imports;
    }
    
    /**
     * Return any class-level annotations that should be documented on this class
     * @return
     */
    public List<String> getClassAnnotations() {
        List<String> annotations = new ArrayList<String>();
        annotations.add(getTarget().getClass().getSimpleName() + "(\"" + getTarget().getName() + "\")");
        return annotations;
    }
    
    /**
     * Construct an array annotation with the supplied name and values
     * @param annotation Annotation name
     * @param values List of values to quote and set in annotation
     * @return String representation of the supplied annotation, e.g. <samp>Annotation({"One", "Two"})</samp>
     */
    protected static String stringAnnotation(String annotation, Collection<String> values) {
        if (values.isEmpty()) {
            return null;
        }
        
        StringBuilder buf = new StringBuilder(annotation + "({");
        Iterator<String> iter = values.iterator();
        while (iter.hasNext()) {
            buf.append("\"")
                .append(iter.next())
                .append("\"");
            
            if (iter.hasNext()) {
                buf.append(", ");
            }
        }
        
        buf.append("})");
        return buf.toString();
    }
    
    public void writeCode(String dirName) throws IOException {
        new ClassGenerator(this.getNameGenerated(), this.getTarget().getFields()).writeCode(dirName, this.getNameGenerated() + ".java");
    }
    
    protected class ClassGenerator extends NetappClassGenerator {
        private String mode;
        private String className;
        private Collection<Field> fields;

        public ClassGenerator(String className, Collection<Field> fields) {
            super(getPackageName(), AutoZapiGenerator.FORMATTER);
            this.className = className;
            this.fields = fields;
            this.mode = AutoZapiTypeGenerator.this.getGeneratorFactory().getProperty(AutoZapiGeneratorFactory.PROPERTY_MODE.getProperty());

            generateClassFile();
        }
        
        /**
         * Generate the class for this type into our class buffer.
         *
         */
        protected void generateClassFile() {
            for (String s : getImports(this.fields)) {
                addImport(s);
            }
            addImport("java.io.Serializable");
            
            addJavadocLine(getDescGenerated());
            
            for (String s : getClassAnnotations()) {
                addLine("@" + s);
            }
            
            addLine("public class " + this.className + " implements Serializable { ");
            increaseIndent();
            addLine("private static final long serialVersionUID = 1L;");
            addLine();
            
            // Field declarations
            for (Field field : this.fields) {
                this.generateFieldDeclaration(field);
            }
            
            // Constructor
            addLine("public " + this.className + "() { }");
            addLine();
            
            for (Field field : this.fields) {
                this.generateFieldAccessors(this.className, field);
            }
            decreaseIndent();
            addLine("}");
        }

        protected final void generateFieldDeclaration(Field field) {
            AutoZapiFieldGenerator<? extends Field> fieldGenerator = (AutoZapiFieldGenerator<? extends Field>) getGeneratorFactory().getFieldGenerator(field);
            addLine("public " + fieldGenerator.getTypeDeclaration() + " " + fieldGenerator.getNameGenerated() + ";");
            addLine();
        }
        
        protected final void generateFieldAccessors(String enclosingClassName, Field field) {
            // Changes from Daniel Holmes
            // For Normal Mode
            //   Basic member get/set methods are generated as normally expected
            // For VMO Mode
            //   VMO plugin model required that Collection members be set as Arrays, not Lists (JavaScript to Java auto converion issue
            //   VMO can handle getters that return Lists, they are converted to Arrays for JavaScript just fine
            //   VMO find bean methods based on attribute names, but reflection doesn't properly distinguish a setter with a List vs Array signature properly,
            //       seems due to reflection indeterminism
            //   AutoZapi runtime requires corresponding set/get with List signature and annotations, actual method names don't matter
            //       but needs to be a get/set pair
            //   So to support all this, for this mode we generate
            //       regular named bean getter for Collection members with no annotations
            //       regular named bean setter for Collection members with no annotations and with Array signature
            //       internal named bean getter for Collection members (just like Normal except for the name)
            //       internal named bean setter for Collection members (just like Normal except for the name)
            //
            // This issue could be dealt with potentially with a VMO oriented subclassing of AutoZapi similar to what Eric did for GWT
            // However this change is internal to the structure of the class file generation (whereas GWT was based on different classes based
            // on input/output field separation, but with the class pattern was the same) and the internal class elements is not fully
            // refactored in a template method pattern which would more easily support generation of specific class methods.
            // Also, this exercise with VMO determined some control aspects for for getter/setter generation, but this doesn't seem complete
            // enough to support the total refactoring an parameterizations needed for cases that might occur beyond VMO.
            // So for now this is handled with the addition of the mode property, but should be refactorable in the future when more time and
            // information is available.

            if (AutoZapiGeneratorFactory.PROPERTY_MODE.getDefaultValue().equals(mode)) {
                generateFieldSetter(field, true, true, true);
                generateFieldGetter(field, true);
            }
            else if (AutoZapiGeneratorFactory.PROPERTY_MODE_VMO.equals(mode)) {
                // normal bean named for VMO
                generateFieldSetter(field, true, false, false);
                generateFieldGetter(field, true);
                // internal named for AutoZapi runtime
                generateFieldSetter(field, false, true, true);
                generateFieldGetter(field, false);
            }
            else {
                // this case should not occur via CLI invocation method, but could in a programmatic use at some point
                // so just in case adding this check.
                throw new RuntimeException("Internal Error: AutoZapi generation mode property not set or value invalid - " + mode);
            }
            
            generateFieldWith(enclosingClassName, field);
        }

        protected final void generateFieldWith(String enclosingClassName, Field field) {
            AutoZapiFieldGenerator<? extends Field> fieldGenerator = (AutoZapiFieldGenerator<? extends Field>) getGeneratorFactory().getFieldGenerator(field);
            String fieldNameGenerated = fieldGenerator.getNameGenerated();

            if (field.isChainable()) {
                String methodName = "with" + fieldGenerator.getMethodSuffix();
                addJavadocLine(fieldGenerator.getDescGenerated());
                addLine("public " + enclosingClassName + " " + methodName + "(" + fieldGenerator.getTypeDeclaration() + " " + fieldNameGenerated + ") {");
                increaseIndent();
                addLine("this." + fieldNameGenerated + " = " + fieldNameGenerated + ";");
                addLine("return this;");
                decreaseIndent();
                addLine("}");
                addLine();
                
                // Add a withFoo(Foo... foo) setter as well.
                if (field.isCollection()) {
                    addImport("java.util.Arrays");
                    addImport("java.util.ArrayList");
                    addJavadocLine(fieldGenerator.getDescGenerated());
                    addLine("public " + enclosingClassName + " " + methodName + "(" + fieldGenerator.getUnqualifiedTypeClass() + "... " + fieldGenerator.getNameGenerated() + ") {");
                    increaseIndent();
                    addLine("this." + fieldNameGenerated + " = new ArrayList<" + fieldGenerator.getUnqualifiedTypeClass() + ">(Arrays.asList(" + fieldNameGenerated + "));");
                    addLine("return this;");
                    decreaseIndent();
                    addLine("}");
                    addLine();
                }
            }
        }

        protected final void generateFieldSetter(Field field, boolean beanNaming, boolean annotate, boolean collectionAsList) {
            AutoZapiFieldGenerator<? extends Field> fieldGenerator = (AutoZapiFieldGenerator<? extends Field>) getGeneratorFactory().getFieldGenerator(field);
            String fieldNameGenerated = fieldGenerator.getNameGenerated();
            // opposite of collectionAsList is to treat it as an array

            String methodName = "set" + fieldGenerator.getMethodSuffix() + (beanNaming ? "" : "Internal");
            addJavadocLine(fieldGenerator.getDescGenerated());

            if (annotate) {
                for (String s : fieldGenerator.getFieldAnnotations()) {
                    addLine("@" + s);
                }
            }

            if (fieldGenerator.getTarget().isCollection() && !collectionAsList) {
                addLine("public void " + methodName + "(" + fieldGenerator.getUnqualifiedTypeClass() + "... " + fieldNameGenerated + ") {");
            }
            else {
                addLine("public void " + methodName + "(" + fieldGenerator.getTypeDeclaration() + " " + fieldNameGenerated + ") {");
            }
            increaseIndent();
            if (fieldGenerator.getTarget().isCollection() && !collectionAsList) {
                addImport("java.util.Arrays");
                addImport("java.util.ArrayList");
                addLine("this." + fieldNameGenerated + " = new ArrayList<" + fieldGenerator.getUnqualifiedTypeClass() + ">(Arrays.asList(" + fieldNameGenerated + "));");
            }
            else {
                addLine("this." + fieldNameGenerated + " = " + fieldNameGenerated + ";");
            }
            decreaseIndent();
            addLine("}");
            addLine();
        }

        protected final void generateFieldGetter(Field field, boolean beanNaming) {
            AutoZapiFieldGenerator<? extends Field> fieldGenerator = (AutoZapiFieldGenerator<? extends Field>) getGeneratorFactory().getFieldGenerator(field);
            String fieldNameGenerated = fieldGenerator.getNameGenerated();

            String methodName;
            if (fieldGenerator.getTypeDeclaration().equals("Boolean")) {
                methodName = "is" + fieldGenerator.getMethodSuffix() + (beanNaming ? "" : "Internal");
            }
            else {
                methodName = "get" + fieldGenerator.getMethodSuffix() + (beanNaming ? "" : "Internal");
            }

            addLine("public " + fieldGenerator.getTypeDeclaration() + " " + methodName + "() {");
            increaseIndent();
            addLine("return this." + fieldNameGenerated + ";");
            decreaseIndent();
            addLine("}");
            addLine();
        }
    }
}