/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/parser/Type.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.parser;

import java.util.Collection;
import java.util.HashSet;
import java.util.Queue;
import java.util.Set;
import java.util.TreeSet;

import org.apache.commons.lang.builder.CompareToBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.zapi.generator.GeneratorFactory;

/**
 * Base class for defined types in zapi, either an Api call or a
 * Typedef struct
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public abstract class Type implements Comparable<Type> {
    private static final NALogger logger = NALogger.getLogger(Type.class);
    
    /** The declared zapi name of this type */
    protected String name;
    
    /** The parser we were created with */
    protected ZapidocParser parser;
    
    /** The category of this type */
    protected Category category;
    
    /** The description on this type */
    protected String desc;
    
    /** The namespace of this type */
    protected Namespace namespace;
    
    /** The fields declared on this type. */
    protected Set<Field> fields = new TreeSet<Field>();
    
    /**
     * Construct a new Type
     * @param category Category of this type
     * @param name zapi name of this type
     */
    protected Type(ZapidocParser parser, String name) {
        this.parser = parser;
        if (parser != null) {
            this.category = parser.getCurrentCategory();
        }
        this.name = name;
    }
    
    /**
     * Get the declared zapi name of this type
     * @return zapi name
     */
    public String getName() {
        return this.name;
    }
    
    /**
     * 
     */
    public Category getCategory() {
        return this.category;
    }
    
    /**
     * Set the description on this type
     * @param desc description of this type
     */
    protected void setDesc(String desc) {
        this.desc = desc;
    }
    
    /**
     * Get the description of this type
     * @return description
     */
    public String getDesc() {
        return this.desc;
    }
    
    /**
     * Set the namespace of this type. It is an error to attempt to
     * redefine the namespace once set.
     * 
     * @param n Name of the enclosing namespace
     * @see #setNamespace(Namespace)
     */
    void setNamespace(String namespaceName) {
        this.setNamespace(parser.getNamespace(namespaceName));
    }
    
    /**
     * Set the namespace of this type. It is an error to attempt to
     * redefine the namespace once set.
     * @param namespace enclosing namespace 
     */
    void setNamespace(Namespace namespace) {
        if (this.namespace != null && this.namespace == namespace) {
            this.parser.appendWarning(MsgKey.ZAPI_GENERATOR_REDEFINE_NAMESPACE_SAME, this.toString(), this.namespace);
            return;
        }
        else if (this.namespace != null) {
            this.parser.appendError(MsgKey.ZAPI_GENERATOR_REDEFINE_NAMESPACE, this.toString(), this.namespace, namespace);
            return;
        }
        this.namespace = namespace;
        this.namespace.addType(this);
        parser.setLastNamespace(namespace);
    }
    
    /**
     * Get the namespace containing this type
     * @return enclosing namespace
     */
    public Namespace getNamespace() {
        return this.namespace;
    }
    
    /**
     * Populate this namespace from the supplied queue of tags, stopping
     * at the first unrecognized tag. Any processed tags are removed from the queue.
     * @param q Queue of tags to process
     */
    abstract void populate(Queue<Tag> q);
    
    /**
     * Return the fields defined by this class
     * @return
     */
    public Collection<Field> getFields() {
        return fields;
    }
    
    /**
     * @return all of the referenced types of my fields
     */
    public Collection<Type> getReferencedTypes() {
        Set<Type> references = new HashSet<Type>();
        for (Field f : fields) {
            Type destType = f.getTypedefType();
            if (destType != null && !destType.isSimple()) {
                references.add(destType);
            }
        }
        return references;
    }
    
    /**
     * @return all of the referenced types of my fields, and of their fields, etc (recursively)
     */
    public Collection<Type> getAllReferencedTypes() {
        Set<Type> references = new HashSet<Type>();
        getAllReferencedTypes(references);
        return references;
    }

    protected void getAllReferencedTypes(Collection<Type> types) {
        for (Field f : fields) {
            Type destType = f.getTypedefType();
            if (destType != null && !destType.isSimple() && !types.contains(destType)) {
                types.add(destType);
                destType.getAllReferencedTypes(types);
            }
        }
    }

    /**
     * Check if this type is an alias for another type
     * @return
     */
    public boolean isAlias() {
        return false;
    }
    
    public boolean isSimple() {
        return false;
    }
    
    /**
     * If this is an alias, then return the root
     * of the alias hierarchy.
     * 
     * If this typedef is not an alias, then return this.
     */
    public Type getRootType() {
        return this;
    }

    protected void addWordReplacements(GeneratorFactory<?, ?, ?, ?> generatorFactory) {
        generatorFactory.getDescriptionGenerator().addWordReplacement(
            this.name,
            generatorFactory.getTypeGenerator(this).getNameGenerated());
        
        for (Field field : getFields()) {
            field.addWordReplacements(generatorFactory);
        }
    }
    
    /**
     * Return a string representation of this type
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return this.getClass().getSimpleName() + "[" + this.getName() + "]";
    }
    
    @Override
    public int hashCode() {
        return new HashCodeBuilder()
            .append(this.namespace)
            .append(this.category)
            .append(this.name)
            .toHashCode();
    }
    
    public int compareTo(Type obj) {
        return new CompareToBuilder()
            .append(this.namespace, obj.namespace)
            .append(this.category, obj.category)
            .append(this.name, obj.name)
            .toComparison();
    }
}
